"""Base Phase Entropy function."""
import numpy as np
import matplotlib.pyplot as mpl

def PhasEn(Sig, K=4, tau=1, Logx=np.exp(1), Norm=True, Plotx=False):
    """PhasEn  estimates the phase entropy of a univariate data sequence.

    .. code-block:: python
    
        Phas = PhasEn(Sig) 
        
    Returns the phase entropy (``Phas``) estimate of the data sequence (``Sig``)
    using the default parameters: angular partitions = 4, time delay = 1, logarithm = natural,
 
    .. code-block:: python
    
        Phas = PhasEn(Sig, keyword = value, ...)
        
    Returns the phase entropy (``Phas``) estimate of the data sequence (``Sig``)  
    using the specified 'keyword' arguments:
        :K:     - Angular partitions (coarse graining), an integer > 1
                  Note: Division of partitions begins along the positive
                        x-axis. As this point is somewhat arbitrary, it is
                        recommended to use even-numbered (preferably
                        multiples of 4) partitions for sake of symmetry. 
        :tau:   - Time Delay, a positive integer
        :Logx:  - Logarithm base, a positive scalar  
        :Norm:  - Normalisation of Phas value, a boolean:
            - [false]  no normalisation
            - [true]   normalises w.r.t. the # partitions ``Log(K)``  (Default)
            
    :Plotx: - When ``Plotx == true``, returns second-order difference plot (default: false)
    
    :See also:
        ``SampEn``, ``ApEn``, ``GridEn``, ``MSEn``, ``SlopEn``, ``CoSiEn``, ``BubbEn``
    
    :References:
        [1] Ashish Rohila and Ambalika Sharma,
            "Phase entropy: a new complexity measure for heart rate
            variability." 
            Physiological measurement
            40.10 (2019): 105006.
    
    """
    
    Sig = np.squeeze(Sig)
   
    assert Sig.shape[0]>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(K,int) and (K > 1), "K:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"  
    assert isinstance(Plotx,bool), "Plotx:     must be boolean - True or False"                          
        
    Yn = Sig[2*tau:] - Sig[tau:-tau]
    Xn = Sig[tau:-tau] - Sig[:-2*tau]    
    with np.errstate(divide='ignore', invalid='ignore'):
        Theta_r = np.arctan(Yn/Xn)   
        Theta_r[np.logical_and((Yn<0),(Xn<0))] += np.pi
        Theta_r[np.logical_and((Yn<0),(Xn>0))] += 2*np.pi
        Theta_r[np.logical_and((Yn>0),(Xn<0))] += np.pi
    
    Limx = np.ceil(np.max(np.abs([Yn, Xn])))
    Angs = np.linspace(0,2*np.pi,K+1)
    Tx = np.zeros((K,len(Theta_r)))
    Si = np.zeros(K)
    
    for n in range(K):
        Temp = np.logical_and((Theta_r > Angs[n]), (Theta_r < Angs[n+1]))
        Tx[n,Temp] = 1
        Si[n] = np.sum(Theta_r[Temp])
            
    Si = Si[Si!=0]
    Phas = -np.sum((Si/np.sum(Si))*(np.log(Si/np.sum(Si))/np.log(Logx)))
    if Norm:
        Phas = Phas/(np.log(K)/np.log(Logx))
    
    if Plotx:

        Tx = Tx.astype(bool)
        Ys = np.sin(Angs)*Limx*np.sqrt(2)
        Xs = np.cos(Angs)*Limx*np.sqrt(2)
        Cols = np.transpose(np.vstack((np.zeros(K), np.tile((np.random.permutation(K)+1)/K,(2,1)))))
        mpl.figure()
        for n in range(K):
            mpl.plot(Xn[Tx[n,:]],Yn[Tx[n,:]],'.',color=tuple(Cols[n,:]))
            mpl.plot(np.vstack((np.zeros(K+1),Xs)),np.vstack((np.zeros(K+1),Ys)),color='m')
        mpl.axis([-Limx, Limx, -Limx, Limx])
        mpl.gca().set_aspect('equal','box')
        mpl.xlabel(r'$X(n +  \tau)  -  X(n)$'), 
        mpl.ylabel(r'$X(n + 2 \tau)  -  X(n + \tau)$')
        mpl.xticks([-Limx, 0, Limx])
        mpl.yticks([-Limx, 0, Limx])
        mpl.show()
    
    return Phas
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base MSobject function."""

def MSobject(EnType='SampEn', **kwargs):
    """MSobject  creates an object to store multiscale entropy parameters.
    
    .. code-block:: python
    
        [Mobj] = MSobject() 
        
    Returns a multiscale entropy object (``Mobj``) based on that originally
    proposed by Costa et al. using the following default  parameters:
    EnType = 'SampEn', embedding dimension = 2, time delay = 1, 
    radius = 0.2*SD(``Sig``), logarithm = natural
     
    .. code-block:: python
    
        [Mobj] = MSobject(EnType)
        
    Returns a multiscale entropy object using the specified entropy method
    (``EnType``) and the default parameters for that entropy method.
    To see the default parameters for a particular entropy method,          
    type:   `help(EnType)`   (e.g.  ``help(SampEn)``)
    
    .. code-block:: python
    
        [Mobj] = MSobject(EnType, keyword = value, ...)
        
    Returns a multiscale entropy object using the specified entropy method
    (``EnType``) and the name/value parameters for that particular method.
    To see the default parameters for a particular entropy method,          
    type:   `help(EnType)`   (e.g.  ``help(SampEn)``)
    
    ``EnType`` can be any of the following (case sensitive) string names:
    
    :Base Entropies: 
        :``'ApEn'``:      - Approximate Entropy
        :``'SampEn'``:    - Sample Entropy
        :``'FuzzEn'``:    - Fuzzy Entropy
        :``'K2En'``:      - Kolmogorov Entropy
        :``'PermEn'``:    - Permutation Entropy	
        :``'CondEn'``:    - Conditional Entropy	
        :``'DistEn'``:    - Distribution Entropy	
        :``'DispEn'``:    - Dispersion Entropy	
        :``'SpecEn'``:    - Spectral Entropy
        :``'SyDyEn'``:    - Symbolic Dynamic Entropy	
        :``'IncrEn'``:    - Increment Entropy	
        :``'CoSiEn'``:    - Cosine Similarity Entropy	
        :``'PhasEn'``:    - Phase Entropy	
        :``'SlopEn'``:    - Slope Entropy
        :``'BubbEn'``:    - Bubble Entropy	
        :``'GridEn'``:    - Grid Distribution Entropy	
        :``'EnofEn'``:    - Entropy of Entropy	
        :``'AttnEn'``:    - Attention Entropy
        :``'DivEn'``:     - Diversity Entropy
        :``'RangEn'``:    - Range Entropy
        
    
    :Cross Entropies:    
        :``'XApEn'``:     - Cross-Approximate Entropy
        :``'XSampEn'``:   - Cross-Sample Entropy
        :``'XFuzzEn'``:   - Cross-Fuzzy Entropy
        :``'XK2En'``:     - Cross-Kolmogorov Entropy
        :``'XPermEn'``:   - Cross-Permutation Entropy
        :``'XCondEn'``:   - Cross-Conditional Entropy (corrected)
        :``'XDistEn'``:   - Cross-Distribution Entropy
        :``'XSpecEn'``:   - Cross-Spectral Entropy
        
     
    :Multivariate Entropies:          
        :``'MvSampEn'``:   - Multivariate Sample Entropy
        :``'MvFuzzEn'``:   - Multivariate Fuzzy Entropy
        :``'MvDispEn'``:   - Multivariate Dispersion Entropy
        :``'MvCoSiEn'``:   - Multivariate Cosine Similarity Entropy
        :``'MvPermEn'``:   - Multivariate Permutation Entropy   

    
    :See also:
        ``MSEn``, ``MvMSEn``, ``cMSEn``, ``cMvMSEn``,  ``rMSEn``, ``hMSEn``, ``XMSEn``, ``rXMSEn``, ``cXMSEn``, ``hXMSEn``    
    """
        
    Chk = ['_ApEn',   '_SampEn', '_FuzzEn', '_K2En',   '_PermEn', '_CondEn', 
           '_DistEn', '_DispEn', '_SyDyEn', '_IncrEn', '_CoSiEn', '_PhasEn', 
           '_SpecEn', '_SlopEn', '_GridEn', '_BubbEn', '_EnofEn', '_AttnEn',
           '_DivEn', '_RangEn', 
           '_XApEn', '_XSampEn', '_XFuzzEn', '_XPermEn', 
           '_XCondEn', '_XDistEn','_XSpecEn', '_XK2En', 
           '_MvSampEn', '_MvFuzzEn', '_MvDispEn', '_MvPermEn','_MvCoSiEn']
    
    assert ('_'+EnType) in Chk, "EnType:      must be a valid entropy function name. \
    See help(EntropyHub.MSobject) for more info."
                
    class MS_Entropy:
        def __init__(self, EnType, X):
            # _temp = __import__(('_'+EnType), globals(), locals(), [EnType], 0)
            
            _temp = __import__(('EntropyHub._'+EnType), fromlist=[EnType])  # USE THIS AS METHOD FOR PUBLISHED PACKAGE
            # _temp = __import__('_'+EnType)        # USE THIS FOR PACKAGE DEVEOPMENT
            
            self.Func = getattr(_temp,EnType) 
            self.Kwargs = X
    
    Mobj = MS_Entropy(EnType, kwargs)   
    return Mobj

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Cross Fuzzy Entropy function."""
import numpy as np
    
def XFuzzEn(*Sig, m=2, tau=1, r=(.2,2.0), Fx= 'default', Logx=np.exp(1)):
    """XFuzzEn  estimates the cross-fuzzy entropy between two univariate data sequences.

    .. code-block:: python
    
        XFuzz, Ps1, Ps2 = XFuzzEn(Sig1,Sig2) 
    
    Returns the cross-fuzzy entropy estimates (``XFuzz``) and the average fuzzy
    distances (``m: Ps1``, ``m+1: Ps2``) for ``m`` = [1,2] estimated for the data sequences
    contained in ``Sig1`` and ``Sig2``, using the default parameters: embedding dimension = 2,
    time delay = 1, fuzzy function (``Fx``) = 'default', 
    fuzzy function parameters (``r``) = (0.2, 2), logarithm = natural
    
    .. code-block:: python
    
        XFuzz, Ps1, Ps2 = XFuzzEn(Sig1, Sig2, keyword = value, ...)
        
    Returns the cross-fuzzy entropy estimates (``XFuzz``) for dimensions = [1, ..., ``m``]
    estimated for the data sequences in ``Sig1`` and ``Sig2`` using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, a positive integer   [default: 2]
        :tau:   - Time Delay, a positive integer            [default: 1]
        :Fx:    - Fuzzy function name, one of the following strings:  
            {``'sigmoid'``, ``'modsampen'``, ``'default'``, ``'gudermannian'``, ``'bell'``, ``'triangular'``, ``'trapezoidal1'``, ``'trapezoidal2'``, ``'z_shaped'``, ``'gaussian'``, ``'constgaussian'``}
        :r:     - Fuzzy function parameters, a 1 element scalar or a 2 element vector of positive values. 
                The ``r`` parameters for each fuzzy  function are defined as follows:   [default: (.2 2)]
                  - default:      [Tuple]
                      * r(1) = divisor of the exponential argument
                      * r(2) = argument exponent (pre-division)  
                  - sigmoid:      [Tuple]
                      * r(1) = divisor of the exponential argument
                      * r(2) = value subtracted from argument (pre-division)
                  - modsampen:    [Tuple]
                      * r(1) = divisor of the exponential argument
                      * r(2) = value subtracted from argument (pre-division)
                  - gudermannian:   
                      * r  = a scalar whose value is the numerator of  argument to gudermannian function: GD(x) = atan(tanh(r/x)). GD(x) is normalised to have a maximum value of 1.
                  - triangular:  
                      * r = a scalar whose value is the threshold (corner point) of the triangular function.
                  - trapezoidal1:  
                      * r = a scalar whose value corresponds to the upper (2r) and lower (r) corner points of the trapezoid.
                  - trapezoidal2:  [Tuple]
                      * r(1) = a value corresponding to the upper corner point of the trapezoid.
                      * r(2) = a value corresponding to the lower corner point of the trapezoid.
                  - z_shaped:  
                      * r = a scalar whose value corresponds to the upper (2r) and lower (r) corner points of the z-shape.
                  - bell:  
                      * r(1) = divisor of the distance value
                      * r(2) = exponent of generalized bell-shaped function
                  - gaussian:  
                      * r = a scalar whose value scales the slope of the Gaussian curve.
                  - constgaussian:  
                      * r = a scalar whose value defines the lower threshod and shape of the Gaussian curve.                    
                  - [DEPRICATED] linear:       
                      r  = an integer value. When r = 0, the argument of the exponential function is 
                      normalised between [0 1]. When r = 1, the minimuum value of the exponential argument is set to 0.    
        :Logx:  - Logarithm base, a positive scalar  [default: natural]
    
    For further information on the keyword arguments, see the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_.
    
    :See also:
        ``FuzzEn``, ``XSampEn``, ``XApEn``, ``FuzzEn2D``, ``XMSEn``, ``MSEn``
    
    :References:
        [1] Hong-Bo Xie, et al.,
            "Cross-fuzzy entropy: A new method to test pattern synchrony of
            bivariate time series." 
            Information Sciences 
            180.9 (2010): 1715-1724.
            
        [2] Hamed Azami, et al.
            "Fuzzy Entropy Metrics for the Analysis of Biomedical Signals: 
            Assessment and Comparison"
            IEEE Access
            7 (2019): 104833-104847
     """
    
    assert len(Sig)<=2,  """Input arguments to be passed as data sequences:
        - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
        - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
    if len(Sig)==1:
        Sig = np.squeeze(Sig)
        assert max(Sig.shape)>10 and min(Sig.shape)==2,  """Input arguments to be passed as data sequences:
            - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
            - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
        if Sig.shape[0] == 2:
            Sig = Sig.transpose()  
        S1 = Sig[:,0]; S2 = Sig[:,1]     
        
    elif len(Sig)==2:
        S1 = np.squeeze(Sig[0])
        S2 = np.squeeze(Sig[1])
        
    N  = S1.shape[0]
    N2 = S2.shape[0]     
    assert N>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"
    assert isinstance(m,int) and (m>0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau>0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) or ((r[0] >= 0) and len(r) ==2), "r:     must be 2 element tuple of positive values"    
    assert Fx.lower() in ['default','sigmoid','modsampen','gudermannian', 'bell', 
                          'gaussian', 'constgaussian', 'triangular' , 
                          'trapezoidal1', 'trapezoidal2', 'z_shaped'] \
            and isinstance(Fx,str), """Fx:    must be one of the following strings -
            'default', 'sigmoid', 'modsampen', 'gudermannian', 'bell', 'z_shaped',
            'triangular', 'trapezoidal1','trapezoidal2','gaussian','constgaussian'"""
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    
    if isinstance(r,tuple) and Fx.lower()=='linear':
        r = 0
        print('Multiple values for r entered. Default value (0) used.') 
    elif isinstance(r,tuple) and Fx.lower()=='gudermannian':
        r = r[0];
        print('Multiple values for r entered. First value used.')      
        
    m = m+1      
    Fun = globals()[Fx.lower()]   
    Sx1 = np.zeros((N,m))
    Sx2 = np.zeros((N2,m))    
    for k in range(m):
        Sx1[:N-k*tau,k] = S1[k*tau::]
        Sx2[:N2-k*tau,k] = S2[k*tau::]
        
    Ps1 = np.zeros(m)
    Ps2 = np.zeros(m-1)
    Ps1[0] = 1    
    for k in range(2,m+1):        
        N1x = N - k*tau
        N2x = N - (k-1)*tau        
        N1y = N2 - k*tau
        N2y = N2 - (k-1)*tau
                
        A = Sx1[:N2x,:k] - np.transpose(np.tile(np.mean(Sx1[:N2x,:k],axis=1),(k,1)))
        B = Sx2[:N2y,:k] - np.transpose(np.tile(np.mean(Sx2[:N2y,:k],axis=1),(k,1)))
        d2 = np.zeros((N2x,N2y))        
                    
        for p in range(N2x):
            Mu2 = np.max(np.abs(A[p,:] - B),axis=1)
            d2[p,:] = Fun(Mu2,r)   
        
        Ps1[k-1] = np.mean(d2[:N1x,:N1y])
        Ps2[k-2] = np.mean(d2)
        
    # m = m+1      
    # Fun = globals()[Fx.lower()]   
    # Sx1 = np.zeros((N,m))
    # Sx2 = np.zeros((N,m))    
    # for k in range(m):
    #     Sx1[:N-k*tau,k] = S1[k*tau::]
    #     Sx2[:N-k*tau,k] = S2[k*tau::]
        
    # Ps1 = np.zeros(m)
    # Ps2 = np.zeros(m-1)
    # Ps1[0] = 1    
    # for k in range(2,m+1):        
    #     N1 = N - k*tau
    #     N2 = N - (k-1)*tau
    #     A = Sx1[:N2,:k] - np.transpose(np.tile(np.mean(Sx1[:N2,:k],axis=1),(k,1)))
    #     B = Sx2[:N2,:k] - np.transpose(np.tile(np.mean(Sx2[:N2,:k],axis=1),(k,1)))
    #     d2 = np.zeros((N2,N2))        
                    
    #     for p in range(N2):
    #         Mu2 = np.max(np.abs(A[p,:] - B),axis=1)
    #         d2[p,:] = Fun(Mu2,r)   
        
    #     Ps1[k-1] = np.mean(d2[:N1,:N1])
    #     Ps2[k-2] = np.mean(d2)
        
    with np.errstate(divide='ignore', invalid='ignore'):
        XFuzz = (np.log(Ps1[:-1]) - np.log(Ps2))/np.log(Logx)
        
    return XFuzz, Ps1, Ps2   

def sigmoid(x,r):
    assert isinstance(r,tuple), 'When Fx = "Sigmoid", r must be a two-element tuple.'
    y = 1/(1 + np.exp((x-r[1])/r[0]))
    return y  
def default(x,r):   
    assert isinstance(r,tuple), 'When Fx = "Default", r must be a two-element tuple.'
    y = np.exp(-(x**r[1])/r[0])
    return y     
def modsampen(x,r):
    assert isinstance(r,tuple), 'When Fx = "Modsampen", r must be a two-element tuple.'
    y = 1/(1 + np.exp((x-r[1])/r[0]));
    return y    
def gudermannian(x,r):
    if r <= 0:
        raise Exception('When Fx = "Gudermannian", r must be a scalar > 0.')
    y = np.arctan(np.tanh(r/x))    
    y = y/np.max(y)    
    return y    
# def linear(x,r):    
#     if r == 0 and x.shape[0]>1:    
#         y = np.exp(-(x - min(x))/np.ptp(x))
#     elif r == 1:
#         y = np.exp(-(x - min(x)))
#     elif r == 0 and x.shape[0]==1:   
#         y = 0;
#     else:
#         print(r)
#         raise Exception('When Fx = "Linear", r must be 0 or 1')
#     return y
def triangular(x, r):
    assert np.size(r)==1, 'When Fx = "Triangular", r must be a scalar > 0.'
    y = 1 - (x/r)
    y[x > r] = 0
    return y
def trapezoidal1(x, r):
    assert np.size(r)==1, 'When Fx = "Trapezoidal1", r must be a scalar > 0.'
    y = np.zeros(np.shape(x))
    y[x <= r*2] = 2 - (x[x <= r*2]/r)
    y[x <= r] = 1
    return y
def trapezoidal2(x, r):
    assert isinstance(r,tuple), 'When Fx = "Trapezoidal2", r must be a two-element tuple.'
    y = np.zeros(np.shape(x))
    y[x <= max(r)] = 1 - (x[x <= max(r)] - min(r))/(max(r)-min(r))
    y[x <= min(r)] = 1
    return y
def z_shaped(x, r):
    assert np.size(r)==1, 'When Fx = "Z_shaped", r must be a scalar > 0.'
    y = np.zeros(np.shape(x))
    y[x <= 2*r] = 2*(((x[x <= 2*r] - 2*r)/r)**2)
    y[x <= 1.5*r] = 1 - (2*(((x[x <= 1.5*r] - r)/r)**2))
    y[x <= r] = 1
    return y
def bell(x, r):
    assert isinstance(r,tuple), 'When Fx = "Bell", r must be a two-element tuple.'
    y = 1/(1 + abs(x/r[0])**(2*r[1]))
    return y
def gaussian(x, r):
    assert np.size(r)==1, 'When Fx = "Gaussian", r must be a scalar > 0.'
    y = np.exp(-((x**2)/(2*(r**2))))
    return y
def constgaussian(x, r):
    assert np.size(r)==1, 'When Fx = "ConstGaussian", r must be a scalar > 0.'
    y = np.ones(np.shape(x))
    y[x > r] = np.exp(-np.log(2)*((x[x > r] - r)/r)**2)
    return y

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Bidimensional Sample Entropy function."""
import numpy as np    

def SampEn2D(Mat, m=None, tau=1, r=None, Logx=np.exp(1), Lock=True):
    """SampEn2D  Estimates the bidimensional sample entropy of a data matrix.

    .. code-block:: python
    
        SE2D, Phi1, Phi2 = SampEn2D(Mat) 
        
    Returns the bidimensional sample entropy estimate (``SE2D``) and the number
    of matched sub-matricess (``m: Phi1``, ``m+1: Phi2``) estimated for the data 
    matrix (Mat) using the default parameters: time delay = 1,
    radius distance threshold = 0.2*SD(``Mat``), logarithm = natural
    matrix template size = [floor(H/10) floor(W/10)]  (where H and W represent
    the height (rows) and width (columns) of the data matrix ``Mat``) 
    * The minimum dimension size of Mat must be > 10.
 
     .. code-block:: python
     
         SE2D, Phi1, Phi2 = SampEn2D(Mat, keyword = value, ...)
    
    Returns the bidimensional sample entropy (``SE2D``) estimates for the data
    matrix (``Mat``) using the specified 'keyword' arguments:
        :m:     - Template submatrix dimensions, an integer scalar (for sub-matrix with same height and width) or a two-element tuple of integers (height, width) with a minimum value > 1. (default: [floor(H/10) floor(W/10)])
        :tau:   - Time Delay, a positive integer > 1   (default: 1)
        :r:     - Distance Threshold, a positive scalar   (default: 0.2*SD(``Mat``))
        :Logx:  - Logarithm base, a positive scalar    (default: natural)        
        :Lock:  - By default, ``SampEn2D`` only permits matrices with a maximum  size of 128 x 128 to prevent RAM overload. 
              e.g. For ``Mat`` = [200 x 200], ``m`` = 3, and ``tau`` = 1, ``SampEn2D`` 
              creates a vector of 753049836 elements. To enable matrices greater than [128 x 128] elements, set ``Lock = False`` (default: True)
              ``CAUTION: unlocking the permitted matrix size may cause memory``
              ``errors that could lead your Python IDE to crash.``
    
    :See also:
        ``SampEn``, ``FuzzEn2D``, ``DistEn2D``, ``XSampEn``, ``MSEn``
    
    :References:
        [1] Luiz Eduardo Virgili Silva, et al.,
            "Two-dimensional sample entropy: Assessing image texture 
            through irregularity." 
            Biomedical Physics & Engineering Express
            2.4 (2016): 045002.
    """
            
    Mat = np.squeeze(Mat)
    NL,NW = Mat.shape    
    if m is None:
        m = np.array(Mat.shape)//10
    if r == None:
        r = 0.2*np.std(Mat)
           
    assert Mat.ndim==2 and min(Mat.shape)>10 , \
    "Mat:   must be a 2D numpy array with height & width > 10"
    assert (isinstance(m,int) and m>1) or (isinstance(m, (np.ndarray,tuple))  
                                           and len(m)==2 and min(m)>1), \
    "m:     must be an integer > 1, or a 2 element tuple of integers > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) and r>0, "r:     must be a positive value"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Lock,bool) and ((Lock==True and max(Mat.shape)<=128) or Lock==False), \
    "Lock:      To prevent memory storage errors, matrix width & length must \
    have <= 128 elements. To estimate SampEn2D for the current matrix (%d x %d) \
    change Lock to False. \
    CAUTION: unlocking the permitted matrix size may cause memory \
    errors that could lead your Python IDE to crash.."%(NW,NL)
    
    if isinstance(m,int):
        mL = int(m); mW = int(m)        
    else:
        mL = int(m[0]); mW = int(m[1])       
                
    NL = NL - mL*tau
    NW = NW - mW*tau
    X = np.zeros((NL*NW,mL+1,mW+1))
    p = 0
    for k in range(NL):        
        for n in range(NW):
            X[p,:,:] = Mat[k:(mL+1)*tau+k:tau,n:(mW+1)*tau+n:tau]
            p += 1           
                        
    if p != NL*NW:
        print('Warning: Potential error with submatrix division.')        
    Ny = int(p*(p-1)/2)
    if Ny > 300000000:
        print('Warning: Number of pairwise distance calculations is ' + str(Ny))
        
    Y1 = np.zeros(p-1, dtype=int)
    Y2 = np.zeros(p-1, dtype=int)
    for k in range(p-1):
        Temp = (np.max(abs(X[k+1:,:-1,:-1] - X[k,:-1,:-1]),axis=(1,2)) < r)
        Y1[k] = sum(Temp)        
        Temp = np.where(Temp==True)[0] + k + 1
        Y2[k] = sum(np.max(abs(X[Temp,:,:] - X[k,:,:]),axis=(1,2)) < r)
     
    Phi1 = sum(Y1)/Ny
    Phi2 = sum(Y2)/Ny
    SE2D = -np.log(Phi2/Phi1)/np.log(Logx)
    
    return SE2D, (Phi1,Phi2)


    # ix = np.vstack((np.repeat(np.arange(NL),NW), np.tile(np.arange(NW),NL))).T
    # T1a, T2a = [],[]
    # T1b, T2b = [],[]
    # Y1 = np.zeros(p-1,dtype=int)
    # Y2 = np.zeros(p-1,dtype=int)
    # for k in range(p-1):
    #     Temp = (np.max(abs(X[k+1:,:-1,:-1] - X[k,:-1,:-1]), axis=(1,2)) < r)
    #     Y1[k] = Temp.sum()        
    #     T1b.extend(np.tile(ix[k].T,(Y1[k],1)))
    #     T2b.extend(ix[k+1:,:][Temp,:])
                    
    #     Temp2 = np.where(Temp)[0] + k + 1
    #     Temp3 = (np.max(abs(X[Temp2,:,:] - X[k,:,:]), axis=(1,2)) < r)
    #     Y2[k] = sum(Temp3)
    #     T1a.extend(np.tile(ix[k].T,(Y2[k],1)))
    #     T2a.extend(ix[Temp2[Temp3],:])
    
    # Phi1 = sum(Y1)/Ny
    # Phi2 = sum(Y2)/Ny
    # SE2D = -np.log(Phi2/Phi1)/np.log(Logx)
    
    # T1a = np.asarray(T1a);     T2a = np.asarray(T2a)
    # T1b = np.asarray(T1b);     T2b = np.asarray(T2b)    
    
    # Ka = 0
    # for k in range(len(T1a)):
    #     Ka += int(((abs(T2a[:,0]- T1a[k,0]) < (mL+1)*tau) + 
    #            (abs(T2a[:,1]- T1a[k,1]) < (mW+1)*tau)).sum())
        
    # Kb = 0
    # for k in range(len(T1b)):
    #     Kb += int(((abs(T2b[:,0]- T1b[k,0]) < mL*tau) + 
    #            (abs(T2b[:,1]- T1b[k,1]) < mW*tau)).sum())
        

    # # TaL = (abs(np.expand_dims(T1a[:,0],axis=1) - np.expand_dims(T2a[:,0],axis=0)) < (mL+1)*tau)
    # # TaW = (abs(np.expand_dims(T1a[:,1],axis=1) - np.expand_dims(T2a[:,1],axis=0)) < (mW+1)*tau)    
    # # Ka = (TaL + TaW).sum()
    
    # # TbL = (abs(np.expand_dims(T1b[:,0],axis=1) - np.expand_dims(T2b[:,0],axis=0)) < mL*tau)
    # # TbW = (abs(np.expand_dims(T1b[:,1],axis=1) - np.expand_dims(T2b[:,1],axis=0)) < mW*tau)    
    # # Kb = (TbL + TbW).sum()
        
    
    # CP = Y1.sum()/Y2.sum()
    # Vcp = (CP*(1-CP)/Y2.sum()) + (Ka - Kb*(CP**2))/(Y2.sum()**2)
    # return SE2D, (Phi1, Phi2), (Vcp, Ka, Kb)


"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Bidimensional Dispersion Entropy function."""
import numpy as np
from scipy.special import ndtr
from scipy.cluster.vq import kmeans2

def DispEn2D(Mat, m=None, tau=1, c=3, Typex='NCDF', Logx=np.exp(1), Norm=False, Lock=True):
    """DispEn2D  Estimates the bidimensional dispersion entropy of a data matrix.
    
    .. code-block:: python
    
        Disp2D, RDE = DispEn2D(Mat) 
        
    Returns the bidimensional dispersion entropy estimate (``Disp2D``) and 
    reverse bidimensional dispersion entropy (``RDE``) estimated for the data 
    matrix (``Mat``) using the default parameters:  time delay = 1, symbols = 3,
    logarithm = natural, data transform = normalised cumulative density function (ncdf)
    matrix template size = [floor(H/10) floor(W/10)]  
    (where H and W represent the height (rows) and width (columns) of the
    data matrix ``Mat``) 
    * The minimum number of rows and columns of Mat must be > 10.
 
    .. code-block:: python
    
        Disp2D, RDE = DistEn2D(Mat, keyword = value, ...)
    
    Returns the bidimensional distribution entropy (``Dist2D``) estimate for 
    the data matrix (Mat) using the specified 'keyword' arguments:
        :m:     - Template submatrix dimensions, an integer scalar 
                (for submatrix with same height and width) or a two-element vector of
                integers [height, width] with a minimum value > 1.
                (default: [floor(H/10) floor(W/10)])
                
        :tau:   - Time Delay, a positive integer  (default: 1)
        :c:     - Number of symbols, an integer > 1
        :Typex: - Typex of data-to-symbol mapping, one of the following strings: 
                {``'linear'``, ``'kmeans'``, ``'ncdf'``, ``'equal'``}
                See the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_ for more info on these transforms.                   
       
        :Logx:  - Logarithm base, a positive scalar
        :Norm:  - Normalisation of ``Disp2D`` and ``RDE`` values, a boolean:
            	* ``False``  no normalisation - default
           	* ``True``  normalises w.r.t # possible vector permutations (c^m).
              
        :Lock:  - By default, ``DispEn2D`` only permits matrices with a maximum size of 128 x 128 to prevent RAM overload. 
                e.g. For ``Mat`` = [200 x 200], ``m`` = 3, and ``tau`` = 1, ``DispEn2D`` 
                creates a vector of 753049836 elements. To enable matrices greater than [128 x 128] elements, set ``Lock = False`` (default: True)
              
                ``CAUTION: unlocking the permitted matrix size may cause memory``
                ``errors that could lead your Python IDE to crash.``
    
    :See also:
        ``DispEn``, ``SampEn2D``, ``FuzzEn2D``, ``DistEn2D``
    
    :References:
        [1] Hamed Azami, et al.,
            "Two-dimensional dispersion entropy: An information-theoretic 
            method for irregularity analysis of images."
            Signal Processing: Image Communication, 
            75 (2019): 178-187.
    
    """
        
    Mat = np.squeeze(Mat)
    NL,NW = Mat.shape    
    if m is None:  m = np.array(Mat.shape)//10
    if Logx == 0:  Logx = np.exp(1)
           
    assert Mat.ndim==2 and min(Mat.shape)>10 , \
    "Mat:   must be a 2D numpy array with height & width > 10"
    assert (isinstance(m,int) and m>1) \
        or (isinstance(m, (np.ndarray,tuple)) and len(m)==2 and min(m)>1), \
    "m:     must be an integer > 1, or a 2 element tuple of integers > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"     
    assert isinstance(c,int) and (c > 1), "c:     must be an integer > 1"
    assert Typex.lower() in ['linear','kmeans','ncdf','equal'], \
            "Typex:    must be one of the following strings - \
            'linear', 'kmeans', 'ncdf', 'equal'"      
    assert isinstance(Norm,bool), "Norm:     must be boolean (true or false)"   
    assert isinstance(Lock,bool) and ((Lock==True and max(Mat.shape)<=128) or Lock==False), \
    "Lock:      To prevent memory storage errors, matrix width & length must \
    have <= 128 elements. To estimate DispEn2D for the current matrix (%d x %d) \
    change Lock to False. \
    CAUTION: unlocking the permitted matrix size may cause memory \
    errors that could lead your Python IDE to crash.."%(NW,NL)
    
    if isinstance(m,int):
        mL = int(m); mW = int(m)        
    else:
        mL = int(m[0]); mW = int(m[1])   
        
    if Typex.lower() == 'linear':
        Zi = np.digitize(Mat,np.arange(np.min(Mat),np.max(Mat),np.ptp(Mat)/c))
        
    elif Typex.lower() == 'kmeans':        
        Clux, Zx = kmeans2(Mat.flatten(), c, iter=200)
        Zx += 1;  xx = np.argsort(Clux) + 1;     Zi = np.zeros(Mat.size);
        for k in range(1,c+1):
            Zi[Zx==xx[k-1]] = k;
        Zi = np.reshape(Zi,Mat.shape)
        
    elif Typex.lower() == 'ncdf':  
        Zx = ndtr((Mat-np.mean(Mat))/np.std(Mat))
        Zi = np.digitize(Zx,np.arange(0,1,1/c))  
        
        del Zx
                        
    elif Typex.lower() == 'equal':
        ix = np.argsort(Mat.flatten(),kind='mergesort')
        xx = np.round(np.arange(0,2*Mat.size,Mat.size/c)).astype(int)
        Zi = np.zeros(Mat.size)
        for k in range(c):
            Zi[ix[xx[k]:xx[k+1]]] = k+1 
            
        Zi = np.reshape(Zi, Mat.shape)
        del ix, xx, k                   
              
    NL = NL - (mL - 1)*tau
    NW = NW - (mW - 1)*tau
    X = np.zeros((NL*NW,mL*mW))
    p = 0
    for k in range(0,NL):        
        for n in range(0,NW):
              X[p,:] = Zi[k:mL*tau+k:tau,n:mW*tau+n:tau].flatten('F')
              p += 1              
    if p != NL*NW:
        print('Warning: Potential error with submatrix division.')        
        
    T  = np.unique(X,axis=0)
    Nx = T.shape[0]
    Counter = np.zeros(Nx)       
    for n in range(Nx):
        Counter[n] = sum(np.any(X - T[n,:],axis=1)==False)

    Ppi = Counter[Counter!= 0]/X.shape[0]
    if np.longdouble(c)**(mL*mW) > np.longdouble(10)**16:
        raise Exception("""RDE cannot be estimated with c = %d and a submatrix of size %d x %d.
        Required floating point precision exceeds 10^16.
        Consider reducing the template submatrix size (m) or the number of symbols (c).""" %(c, mL, mW))
    
    # RDE = sum(Ppi**2) - (1/Nx)
    RDE = sum((Ppi - (1/(c**(mL*mW))))**2); 
    Disp2D = -sum(Ppi*np.log(Ppi)/np.log(Logx))    
    
    if round(sum(Ppi)) != 1:
        print('Potential error calculating probabilities')    
        
    if Norm:
        Disp2D = Disp2D/(np.log(c**(mL*mW))/np.log(Logx))
        RDE = RDE/(1 - (1/(c**(mL*mW))))     
        
    return Disp2D, RDE

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""

    
    
    



"Base Multivariate Multiscale Entropy function."
import numpy as np   
from matplotlib.pyplot import figure, axes, show
from copy import deepcopy
# from scipy.stats import moment

def MvMSEn(Data, Mbjx, Scales=3, Methodx='coarse', Plotx=False):
    """MvMSEn  Returns the multivariate multiscale entropy of a multivariate dataset.
    
    .. code-block:: python
    
        MSx,CI = MvMSEn(Data, Mobj) 
    
    Returns a vector of multivariate multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data sequences in ``Data`` using the parameters specified 
    by the multiscale object (``Mobj``) over 3 temporal scales with coarse-
    graining (default). 
    
    .. caution::
 
        By default, the ``MvSampEn`` and ``MvFuzzEn`` multivariate entropy algorithms
        estimate entropy values using the "full"  method by comparing delay vectors 
        across all possible ``m+1`` expansions of the embedding space as applied in [1].
        These methods are **not** lower-bounded to 0, like most entropy algorithms,
        so ``MvMSEn`` may return negative entropy values if the base multivariate 
        entropy function is ``MvSampEn`` and ``MvFuzzEn``, even for stochastic processes...
    
    
    .. code-block:: python
    
        MSx,CI = MvMSEn(Data, Mobj, keyword = value, ...)
        
    Returns a vector of multivariate multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data sequences in ``Data`` using the parameters specified by
    the multiscale object (``Mobj``) and the following ``keyword`` arguments:
        :Scales:   - Number of temporal scales, an integer > 1   (default: 3)
        :Methodx:  - Graining method, one of the following: [default: ``'coarse'``]  {``'coarse'``, ``'modified'``, ``'generalized'``}                        
        :Plotx:    - When ``Plotx == True``, returns a plot of the entropy value at each time scale (i.e. the multiscale entropy curve) [default: False]
    
    For further info on these graining procedures see the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_.
    
    :See also:
        ``MSobject``, ``cMvMSEn``, ``MvFuzzEn``, ``MvSampEn``, ``MvPermEn``, ``MvCoSiEn``,  ``MvDispEn``
    
    :References:
        [1] Ahmed Mosabber Uddin, Danilo P. Mandic
             "Multivariate multiscale entropy analysis."
             IEEE signal processing letters 19.2 (2011): 91-94.
        
        [2] Madalena Costa, Ary Goldberger, and C-K. Peng,
             "Multiscale entropy analysis of complex physiologic time series."
             Physical review letters
             89.6 (2002): 068102.
        
        [3] Vadim V. Nikulin, and Tom Brismar,
             "Comment on “Multiscale entropy analysis of complex physiologic
             time series”." 
             Physical Review Letters 
             92.8 (2004): 089803.
        
        [4] Madalena Costa, Ary L. Goldberger, and C-K. Peng. 
             "Costa, Goldberger, and Peng reply." 
             Physical Review Letters
             92.8 (2004): 089804.
        
        [5] Madalena Costa, Ary L. Goldberger and C-K. Peng,
             "Multiscale entropy analysis of biological signals." 
             Physical review E 
             71.2 (2005): 021906.
        
        [6] Ranjit A. Thuraisingham and Georg A. Gottwald,
             "On multiscale entropy analysis for physiological data."
             Physica A: Statistical Mechanics and its Applications
             366 (2006): 323-332.
        
        [7] Ahmed Mosabber Uddin, Danilo P. Mandic
             "Multivariate multiscale entropy: A tool for complexity
             analysis of multichannel data."
             Physical Review E 84.6 (2011): 061918.
    
    """
    
    Mobj = deepcopy(Mbjx)
    Chk = ['coarse','modified','generalized']    
    
    Data = np.squeeze(Data)
    assert Data.shape[0]>10 and Data.ndim==2 and Data.shape[1]>1,  "Data:   must be an NxM numpy matrix where N>10 and M>1"
    N, Dn = Data.shape     
    
    assert isinstance(Mobj,object), "Mobj:  must be a multiscale entropy \
                        object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert Methodx.lower() in Chk, "Methodx:  must be one of the following string names- \
                        'coarse', 'modified' , 'generalized'"    
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
    assert Mobj.Func.__name__[:2] == 'Mv', ("Base entropy estimator must be a multivariate entropy method." 
    "To perform univariate multiscale entropy estimation, use MSEn().")
       
    Func2 = globals()[Methodx.lower()]
    MSx = np.zeros(Scales)
    
    for T in range(1,Scales+1):
        print(' .', end='')
        Temp = Func2(Data, T, Dn)  
        Tempx = Mobj.Func(Temp,**Mobj.Kwargs)[0]                                        
        MSx[T-1] = np.mean(Tempx)
            
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')

    if Plotx:
       figure()
       ax1 = axes()   
       ax1.plot(np.arange(1,Scales+1), MSx, color=(8/255, 63/255, 77/255), linewidth=3)
       ax1.scatter(np.arange(1,Scales+1), MSx, 60, color=(1,0,1))
       ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_title('Multivariate Multiscale %s (%s-graining method)'%(Mobj.Func.__name__[2:], Methodx), 
                     fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))      
       show()
    
    return MSx, CI


def coarse(Z, sx, Dn):
    Ns = Z.shape[0]//sx
    Y = np.zeros((Ns,Dn))
    for k in range(Dn):
        Y[:,k] = np.mean(np.reshape(Z[:sx*Ns,k],(Ns,sx)),axis=1)
    return Y

def modified(Z, sx, Dn):       
    Y = np.zeros((Z.shape[0]-sx+1,Dn))
    for k in range(Dn):
        Y[:,k] = np.convolve(Z[:,k],np.ones(sx),'valid')/sx
    return Y 

def generalized(Z, sx, Dn):
    Ns = Z.shape[0]//sx
    Y = np.zeros((Ns,Dn))
    for k in range(Dn):
        Y[:,k] = np.var(np.reshape(Z[:sx*Ns, k],(Ns,sx)), axis=1)
    return Y
    

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Cross Sample Entropy function."""
import numpy as np 

def XSampEn(*Sig, m=2, tau=1, r=None, Logx=np.exp(1), Vcp=False):
    """XSampEn  Estimates the cross-sample entropy between two univariate data sequences.
    
    .. code-block:: python
    
        XSamp, A, B = XSampEn(Sig1, Sig2) 
        
    Returns the cross-sample entropy estimates (``XSamp``) and the number of 
    matched vectors (``m: B``, ``m+1: A``) for ``m`` = [0,1,2] estimated for the two 
    univariate data sequences contained in ``Sig1`` and ``Sig2`` using the default parameters:
    embedding dimension = 2, time delay = 1, radius = 0.2*SDpooled(``Sig1``,``Sig2``),
    logarithm = natural
    
    .. code-block:: python
    
        XSamp, A, B, (Vcp, Ka, Kb) = XSampEn(Sig1, Sig2, ..., Vcp = True) 
    
    If ``Vcp == True``, an additional tuple ``(Vcp, Ka, Kb)`` is returned with    
    the cross-sample entropy estimates (``XSamp``) and the number of matched state
    vectors (``m: B``, ``m+1: A``). ``(Vcp, Ka, Kb)``  contains the variance of the conditional
    probabilities (``Vcp``, i.e. CP = A/B),  and the number of **overlapping**
    matching vector pairs of lengths m+1 (``Ka``) and m (``Kb``),
    respectively.  Note ``Vcp`` is undefined for the zeroth embedding dimension (m = 0) 
    and due to computational demand, **will take substantially more time to return function outputs.**
    See Appendix B in [2] for more info.
 
    .. code-block:: python 
 
         XSamp, A, B = XSampEn(Sig1, Sig2, keyword = value, ...)
                  
    Returns the cross-sample entropy estimates (``XSamp``) for dimensions [0,1,..., ``m``]
    estimated between the data sequences ``Sig1`` and ``Sig2`` using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, a positive integer  [default: 2]
        :tau:   - Time Delay, a positive integer         [default: 1]
        :r:     - Radius, a positive scalar              [default: 0.2*SDpooled(``Sig1``,``Sig2``)]
        :Logx:  - Logarithm base, a positive scalar      [default: natural]
 
    :See also:
        ``XFuzzEn``, ``XApEn``, ``SampEn``, ``SampEn2D``, ``XMSEn``, ``ApEn``
    
    :References:
        [1] Joshua S Richman and J. Randall Moorman. 
            "Physiological time-series analysis using approximate entropy
            and sample entropy." 
            American Journal of Physiology-Heart and Circulatory Physiology
            (2000)
        
        [2]  Douglas E Lake, Joshua S Richman, M.P. Griffin, J. Randall Moorman
            "Sample entropy analysis of neonatal heart rate variability."
            American Journal of Physiology-Regulatory, Integrative and Comparative Physiology
            283, no. 3 (2002): R789-R797.
    
    """
    
    assert len(Sig)<=2,  """Input arguments to be passed as data sequences:
        - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
        - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
    if len(Sig)==1:
        Sig = np.squeeze(Sig)
        assert max(Sig.shape)>10 and min(Sig.shape)==2,  """Input arguments to be passed as data sequences:
            - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
            - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
        if Sig.shape[0] == 2:
            Sig = Sig.transpose()  
        S1 = Sig[:,0]; S2 = Sig[:,1]     
        
    elif len(Sig)==2:
        S1 = np.squeeze(Sig[0])
        S2 = np.squeeze(Sig[1])
        
    N  = S1.shape[0]
    N2 = S2.shape[0]
    if r is None:
        r = 0.2*np.sqrt((np.var(S1)*(N-1) + np.var(S2)*(N2-1))/(N+N2-1))
     
    assert N>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"
    assert isinstance(m,int) and (m > 0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) and r>=0, "r:     must be a positive value"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Vcp,bool), "Vcp:     must be a Boolean"    

    M = np.hstack((m*np.ones(N-m*tau), np.repeat(np.arange(m-1,0,-1),tau)))   
    Counter = 1*(abs(np.expand_dims(S1,axis=1) - np.expand_dims(S2,axis=0))<= r)
    A = np.zeros(m+1)
    B = np.zeros(m+1)
    A[0] = np.sum(Counter)
    B[0] = N*N2
    
    for n in range(M.shape[0]):
        ix = np.where(Counter[n, :] == 1)[0]        
        for k in range(1,int(M[n]+1)):              
            ix = ix[ix + (k*tau) < N2]
            if not len(ix):    break  
            p1 = np.tile(S1[n: n+1+(tau*k):tau], (ix.shape[0], 1))                       
            p2 = S2[np.expand_dims(ix,axis=1) + np.arange(0,(k*tau)+1,tau)]
            ix = ix[np.amax(abs(p1 - p2), axis=1) <= r] 
            Counter[n, ix] += 1
    
    for k in range(1, m+1):
        A[k] = np.sum(Counter > k)
        B[k] = np.sum(Counter >= k)
    
    with np.errstate(divide='ignore', invalid='ignore'):
        XSamp = -np.log(A/B)/np.log(Logx)
 
    
    if Vcp:
        T1,T2 = np.expand_dims(np.where(Counter>m),axis=1)
        Ka = np.triu(abs(T1-T1.T)<=m*tau,1) + np.triu(abs(T2-T2.T)<=m*tau,1) 

        T1,T2 = np.expand_dims(np.where(Counter[:,:-m*tau]>=m),axis=1)
        Kb = np.triu(abs(T1-T1.T)<=(m-1)*tau,1) + np.triu(abs(T2-T2.T)<=(m-1)*tau,1) 
                    
        Ka = np.sum(Ka)
        Kb = np.sum(Kb)
        CP = A[-1]/B[-1]
        Vcp = (CP*(1-CP)/B[-1]) + (Ka - Kb*(CP**2))/(B[-1]**2)

        return XSamp, A, B, (Vcp, Ka, Kb)
    
    else:
        return XSamp, A, B     


"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Symbolic Dynamic Entropy function."""
import numpy as np
from scipy.cluster.vq import kmeans2

def  SyDyEn(Sig, m=2, tau=1, c=3, Typex='MEP', Logx=np.exp(1), Norm=True):
    """SyDyEn  estimates the symbolic dynamic entropy of a univariate data sequence.

    .. code-block:: python
    
        SyDy, Zt = SyDyEn(Sig) 
        
    Returns the symbolic dynamic entropy (``SyDy``) and the symbolic sequence
    (``Zt``) of the data sequence (``Sig``) using the default parameters: 
    embedding dimension = 2, time delay = 1, symbols = 3, logarithm = natural,
    symbolic partition type = maximum entropy partitioning (``'MEP'``), 
    normalisation = normalises w.r.t # possible vector permutations (``c^m``) 
  
    .. code-block:: python
    
        SyDy, Zt = SyDyEn(Sig, keyword = value, ...)
        
    Returns the symbolic dynamic entropy (``SyDy``) and the symbolic sequence
    (``Zt``) of the data sequence (``Sig``) using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, a positive integer
        :tau:   - Time Delay, a positive integer
        :c:     - Number of symbols, an integer > 1
        :Typex: - Type of symbolic sequence partitioning, one of the following:
         {``'linear'``, ``'uniform'``, ``'MEP'`` (default), ``'kmeans'``} 
        :'Logx:  - Logarithm base, a positive scalar  
        :Norm:  - Normalisation of SyDyEn value, a boolean:
            [False]  no normalisation 
            [True]   normalises w.r.t # possible dispersion patterns (``c^m+1``) - default
  
    See the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_ for more info on these parameters.
    
    :See also:
        ``DispEn``, ``PermEn``, ``CondEn``, ``SampEn``, ``MSEn``
    
    :References:
        [1] Yongbo Li, et al.,
            "A fault diagnosis scheme for planetary gearboxes using 
            modified multi-scale symbolic dynamic entropy and mRMR feature 
            selection." 
            Mechanical Systems and Signal Processing 
            91 (2017): 295-312. 
    
        [2] Jian Wang, et al.,
            "Fault feature extraction for multiple electrical faults of 
            aviation electro-mechanical actuator based on symbolic dynamics
            entropy." 
            IEEE International Conference on Signal Processing, 
            Communications and Computing (ICSPCC), 2015.
    
        [3] Venkatesh Rajagopalan and Asok Ray,
            "Symbolic time series analysis via wavelet-based partitioning."
            Signal processing 
            86.11 (2006): 3309-3320.
    
    """
    
    Sig = np.squeeze(Sig)
    N = Sig.shape[0]
            
    assert N>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(c,int) and (c > 1), "c:     must be an integer > 1"
    assert Typex.lower() in ['linear','uniform','mep','kmeans'], \
            "Typex:   must be one of the following strings - \
            'linear', 'uniform', 'MEP', 'kmeans'" 
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"  
    
    Nx = N-((m-1)*tau)
    if Typex.lower() == 'linear':
        Zt = np.digitize(Sig, np.arange(min(Sig),max(Sig),np.ptp(Sig)/c))
        
    elif Typex.lower() == 'uniform':
        Zt = np.zeros(Sig.shape)
        Ix = np.argsort(Sig)
        # z = np.digitize(np.arange(1,N+1),np.arange(1,N+1,round(N/c)))
        z = np.digitize(np.arange(N),np.arange(0,2*N,N/c))
        Zt[Ix] = z
        
    elif Typex.lower()=='kmeans':
        
        Clux, Zi = kmeans2(Sig, c, iter=200)
        Zi += 1;  xx = np.argsort(Clux) + 1;     Zt = np.zeros(N);
        for k in range(1,c+1):
            Zt[Zi==xx[k-1]] = k;        
        
    else:        # MEP method is default
        Tx = np.sort(Sig)
        Temp = np.hstack((0,np.ceil(np.arange(1,c)*N/c)-1))
        Zt = np.digitize(Sig, Tx[Temp.astype(int)])
        
    Zm = np.zeros((Nx,m))
    for n in range(m):
        Zm[:,n] = Zt[n*tau:Nx+(n*tau)]
    
    T = np.unique(Zm,axis=0)
    Counter = np.zeros(T.shape[0])
    Counter2 = np.zeros((T.shape[0],c))
    Bins = np.arange(0.5,c+1.5,1)
    for n in range(T.shape[0]):
        Ordx = (np.any(Zm - T[n,:],axis=1)==0)  
        Counter[n] = sum(Ordx)/Nx
        Temp = Zm[np.hstack((np.zeros(m*tau,dtype=bool), Ordx[:-(m*tau)])),0]
        Counter2[n,:], _ = np.histogram(Temp,Bins)

    Temp = np.sum(Counter2,axis=1)
    Counter2[Temp>0,:] = Counter2[Temp>0,:]/np.tile(Temp[Temp>0],(c,1)).transpose()    
    Counter2[np.isnan(Counter2)] = 0    
    
    with np.errstate(divide='ignore'):
        P1 = -sum(Counter*np.log(Counter)/np.log(Logx))
        P2 = np.log(np.tile(Counter,(c,1)).transpose()*Counter2)/np.log(Logx)
    P2[np.isinf(P2)] = 0 
    SyDy = P1 - sum(Counter*np.sum(P2,axis=1))
    
    if round(sum(Counter),4) != 1 or max(np.sum(Counter2,axis=1)) != 1:
        print('Potential Error calculating probabilities')
    if Norm:
        SyDy = SyDy/(np.log(c**(m+1))/np.log(Logx))
    
    return SyDy, Zt

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Bubble Entropy function."""
import numpy as np

def BubbEn(Sig, m=2, tau=1, Logx=np.exp(1)):
    """BubbEn  estimates the bubble entropy of a univariate data sequence.
 
    .. code-block:: python
    
        Bubb, H = BubbEn(Sig)
    
    Returns the bubble entropy (``Bubb``) and the conditional Rényi entropy (``H``)
    estimate from the data sequence (``Sig``) using the default parameters: 
    embedding dimension = 2, time delay = 1, logarithm = natural 
  
    .. code-block:: python
        
        Bubb, H = BubbEn(Sig, keyword = value, ...)
    
    Returns the bubble entropy (``Bubb``) estimated from the data sequence (``Sig``)  
    using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, an integer > 1
            BubbEn returns estimates for each dimension [2, ..., ``m``]   
        :tau:   - Time Delay, a positive integer
        :Logx:  - Logarithm base, a positive scalar
 
    :See also:
        ``PhasEn``, ``MSEn``

    :References:
        [1] George Manis, M.D. Aktaruzzaman and Roberto Sassi,
            "Bubble entropy: An entropy almost free of parameters."
            IEEE Transactions on Biomedical Engineering
            64.11 (2017): 2711-2718.
    
    """  

    Sig = np.squeeze(Sig)
    N = Sig.shape[0]
    
    assert N>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 1), "m:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
                
    Sx = np.zeros((N,m+1))
    H = np.zeros(m+1)
    Sx[:,0] = Sig    
    for k in range(2,m+2):
        
        Sx[0:N-(k-1)*tau,k-1] = Sig[(k-1)*tau:N]
        temp = np.copy(Sx[0:N-(k-1)*tau,0:k])
        Swapx, _ = BubbSort(temp)
        _, Locs = np.unique(Swapx, return_counts=True)
        p = Locs/(N-(k-1)*tau)
        H[k-1] = -np.log(np.sum(p**2))/np.log(Logx)
        if np.round(np.sum(p)) != 1:
            raise Exception('Potential error in detected swap number')
        del Swapx, p, Locs, temp
         
    with np.errstate(divide='ignore', invalid='ignore'):
        BE = np.diff(H)/((np.log(np.arange(2,m+2)/np.arange(m)))/np.log(Logx))
    
    BE = BE[1:]
    return BE, H

def BubbSort(Data):
    x, N2 = np.shape(Data)
    swaps = np.zeros(x)
    for y in range(x):
        for t in range(N2-1):
            for kk in range(N2-t-1):
                if Data[y,kk] > Data[y,kk+1]:
                    Data[y,kk], Data[y,kk+1] = Data[y,kk+1], Data[y,kk]
                    swaps[y] += 1
    bsorted = Data
    return swaps, bsorted   
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base cross-Kolmogorov Entropy function."""
import numpy as np

def XK2En(*Sig, m=2, tau=1 , r=None, Logx=np.exp(1)):
    """XK2En  estimates the cross-Kolmogorov entropy between two univariate  data sequences.

    .. code-block:: python
    
        XK2, Ci = XK2En(Sig1, Sig2) 
        
    Returns the cross-Kolmogorov entropy estimates (``XK2``) and the correlation
    integrals (``Ci``) for ``m`` = [1, 2] estimated between the data sequences 
    contained in ``Sig1`` and``Sig2`` using the default parameters: 
    embedding dimension = 2, time delay = 1, 
    distance threshold (``r``) = 0.2*SDpooled(``Sig1``,``Sig2``), logarithm = natural
 
    .. code-block:: python
    
        XK2, Ci = XK2En(Sig1,Sig2, keyword = value, ...)
        
    Returns the cross-Kolmogorov entropy estimates (``XK2``) estimated between
    the data sequences contained in ``Sig1`` and ``Sig2`` using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, a positive integer [default: 2]
        :tau:   - Time Delay, a positive integer         [default: 1]
        :r:     - Radius Distance Threshold, a positive scalar [default: 0.2*SDpooled(``Sig1``,``Sig2``),]
        :Logx:  - Logarithm base, a positive scalar      [default: natural]
     
    :See also:
        ``XSampEn``, ``XFuzzEn``, ``XApEn``, ``K2En``, ``XMSEn``, ``XDistEn``
    
    :References:
        [1] Matthew W. Flood,
            "XK2En - EntropyHub Project"
            (2021) https://github.com/MattWillFlood/EntropyHub
    
    """
 
    assert len(Sig)<=2,  """Input arguments to be passed as data sequences:
        - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
        - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
    if len(Sig)==1:
        Sig = np.squeeze(Sig)
        assert max(Sig.shape)>10 and min(Sig.shape)==2,  """Input arguments to be passed as data sequences:
            - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
            - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
        if Sig.shape[0] == 2:
            Sig = Sig.transpose()  
        S1 = Sig[:,0]; S2 = Sig[:,1]     
        
    elif len(Sig)==2:
        S1 = np.squeeze(Sig[0])
        S2 = np.squeeze(Sig[1])
        
    N  = S1.shape[0]
    N2 = S2.shape[0]
    if r is None:
        r = 0.2*np.sqrt((np.var(S1)*(N-1) + np.var(S2)*(N2-1))/(N+N2-1))
    
    assert N>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"
    assert isinstance(m,int) and (m > 0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) and r>0, "r:     must be a positive value"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"

    m = m+1
    Z1 = np.zeros((N,m))  
    Z2 = np.zeros((N2,m))  
    Ci = np.zeros(m)
    for n in range(m):
        Nx = N - n*tau
        Z1[:Nx,n] = S1[n*tau:]
        Ny = N2 - n*tau
        Z2[:Ny,n] = S2[n*tau:]            
        Norm = np.zeros((Nx,Ny))             
        for k in range(Nx):
            Temp = np.tile(Z1[k,:n+1],(Ny,1)) - Z2[:Ny,:n+1]
            Norm[k,:] = np.linalg.norm(Temp,axis=1)                      
        Ci[n] = np.mean(Norm[:] < r)
               
    # m = m+1
    # Z1 = np.zeros((N,m))  
    # Z2 = np.zeros((N,m))  
    # Ci = np.zeros(m)
    # for n in range(m):
    #     N2 = N - n*tau
    #     Z1[:N2,n] = Sig[n*tau:,0]
    #     Z2[:N2,n] = Sig[n*tau:,1]            
    #     Norm = np.zeros((N2,N2))             
    #     for k in range(N2):
    #         Temp = np.tile(Z1[k,:n+1],(N2,1)) - Z2[:N2,:n+1]
    #         Norm[k,:] = np.linalg.norm(Temp,axis=1)                      
    #     Ci[n] = np.mean(Norm[:] < r)
    
    with np.errstate(divide='ignore', invalid='ignore'):
        XK2 = (np.log(Ci[:-1]/Ci[1:])/np.log(Logx))/tau
    XK2[np.isinf(XK2)] = np.NaN    
    return  XK2, Ci

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Cross Spectral Entropy function."""
import numpy as np  

def XSpecEn(*Sig, N=None, Freqs=(0,1), Logx=np.exp(1), Norm=True):
    """XSpecEn  estimates the cross-spectral entropy between two univariate data sequences.

    .. code-block:: python

        XSpec, BandEn = XSpecEn(Sig1, Sig2) 
        
    Returns the cross-spectral entropy estimate (``XSpec``) of the full cross-
    spectrum and the within-band entropy (``BandEn``) estimated for the data 
    sequences contained in ``Sig1`` and ``Sig2`` using the default  parameters: 
    N-point FFT = 2 * max length of ``Sig1``/``Sig2``, 
    normalised band edge frequencies = [0 1], logarithm = base 2, 
    normalisation = w.r.t # of spectrum/band frequency  values.
 
    .. code-block:: python
    
        XSpec, BandEn = XSpecEn(Sig1, Sig2, keyword = value, ...)
        
    Returns the cross-spectral entropy (``XSpec``) and the within-band entropy 
    (``BandEn``) estimate for the data sequences contained in ``Sig1`` and ``Sig2`` 
    using the  following specified 'keyword' arguments:
        :N:     - Resolution of spectrum (N-point FFT), an integer > 1
        :Freqs: - Normalised and edge frequencies, a scalar in range [0 1]  where 1 corresponds to the Nyquist frequency (Fs/2).
              * Note: When no band frequencies are entered, ``BandEn == SpecEn``
        :Logx:  - Logarithm base, a positive scalar     [default: natural]
        :Norm:  - Normalisation of ``XSpec`` value, one of the following integers:
            [false]  no normalisation.
            [true]   normalises w.r.t # of frequency values within the spectrum/band   [default]
    
    See the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_ for more info.
 
    
    :See also:
        ``SpecEn``, ``fft``, ``XDistEn``, ``periodogram``, ``XSampEn``, ``XApEn``
     
    :References:
        [1] Matthew W. Flood,
            "XSpecEn - EntropyHub Project"
            (2021) https://github.com/MattWillFlood/EntropyHub
    
    """


    assert len(Sig)<=2,  """Input arguments to be passed as data sequences:
        - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
        - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
    if len(Sig)==1:
        Sig = np.squeeze(Sig)
        assert max(Sig.shape)>10 and min(Sig.shape)==2,  """Input arguments to be passed as data sequences:
            - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
            - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
        if Sig.shape[0] == 2:
            Sig = Sig.transpose()  
        S1 = Sig[:,0]; S2 = Sig[:,1]     
        
    elif len(Sig)==2:
        S1 = np.squeeze(Sig[0])
        S2 = np.squeeze(Sig[1])
        
    if N is None:
        N = 2*max(len(S1),len(S2)) + 1
       
    assert min(len(S1),len(S2))>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"
    assert N>1 and isinstance(N,int), "N:   must be an integer > 1"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"       
    assert isinstance(Freqs,tuple) and len(Freqs)==2 and Freqs[0]>=0 and Freqs[1]<=1,\
            "Freq:    must be a two element tuple with values in range [0 1]. \
                      The values must be in increasing order."      
              
    Fx = int(np.ceil(N/2))
    Freqs = np.round(np.array(Freqs)*Fx).astype(int)-1
    Freqs[Freqs==-1] = 0
    
    if Freqs[0] > Freqs[1]:
        raise Exception('Lower band frequency must come first.')
    elif np.diff(Freqs) < 1:
        raise Exception('Spectrum resoution too low to determine bandwidth.')
    elif min(Freqs)<0 or max(Freqs)> Fx:
        raise Exception('Freqs must be normalized w.r.t sampling frequency [0 1].')  
        
    Pt = abs(np.fft.fft(np.convolve(S1,S2),N))
    Pxx = Pt[:Fx]/sum(Pt[:Fx])
    XSpec = -sum(Pxx*np.log(Pxx))/np.log(Logx)
    Pband = Pxx[Freqs[0]:Freqs[1]+1]/sum(Pxx[Freqs[0]:Freqs[1]+1])
    BandEn = -sum(Pband*np.log(Pband))/np.log(Logx)
    
    if Norm:
        XSpec = XSpec/(np.log(Fx)/np.log(Logx))
        BandEn = BandEn/(np.log(np.diff(Freqs)[0]+1)/np.log(Logx))
    return XSpec, BandEn

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


import requests
import numpy as np

def ExampleData(SigName):
    """Imports sample data time series and matrices.
    
    .. code-block:: python
    
        Data = ExampleData(SigName) 
    
    Returns sample datasets (time series or matrices) with specific properties 
    that are commonly used as benchmarks for assessing the performance of various
    entropy methods.
    The datasets returned by ExampleData() are used in the examples provided in
    documentation on www.EntropyHub.xyz and elsewhere.
    ***Note*** ExampleData() requires an internet connection to download and import the required datasets!
    
    ``Data`` is the sample dataset imported corresponding to the string input ``SigName``
    which can be one of the following:
        :SigName:   - One of the following strings:                           
                    :``uniform``:          - uniformly distributed random number sequence in range [0 1], N = 5000
                    :``randintegers``:     - randomly distributed integer sequence in range [1 8], N = 4096
                    :``gaussian``:         - normally distributed number sequence [mean: 0, SD: 1], N = 5000
                    :``henon``:            - X and Y components of the Henon attractor [alpha: 1.4, beta: .3, Xo = 0, Yo = 0], N = 4500
                    :``lorenz``:           - X, Y, and Z components of the Lorenz attractor [sigma: 10, beta: 8/3, rho: 28, Xo = 10, Yo = 20, Zo = 10], N = 5917
                    :``chirp``:            - chirp signal (f0 = .01, t1 = 4000, f1 = .025), N = 5000
                    :``uniform2``:         - two uniformly distributed random number sequences in range [0,1], N = 4096
                    :``gaussian2``:        - two normally distributed number sequences [mean: 0, SD: 1], N = 3000
                    :``randintegers2``:    - two uniformly distributed pseudorandom integer sequences in range [1 8], N = 3000
                    :``uniform_Mat``:      - matrix of uniformly distributed random numbers in range [0 1], N = 50 x 50
                    :``gaussian_Mat``:     - matrix of normally distributed numbers [mean: 0, SD: 1], N = 60 x 120
                    :``randintegers_Mat``: - matrix of randomly distributed integers in range [1 8], N = 88 x 88
                    :``mandelbrot_Mat``:   - matrix representing a Mandelbrot fractal image with values in range [0 255], N = 92 x 115
                    :``entropyhub_Mat``:   - matrix representing the EntropyHub logo with values in range [0 255], N = 127 x 95
                         
    For further info on these graining procedures see the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_.
        
    """
    
    Chk = ["uniform","uniform2","randintegers2","randintegers",
        "henon","chirp","gaussian","gaussian2","lorenz",
        "uniform_Mat", "gaussian_Mat", "entropyhub_Mat",
        "mandelbrot_Mat","randintegers_Mat"]
    
    if SigName not in Chk:
        raise Exception("SigName must be one of the following:\n%s" %Chk)
        
    url = r"https://raw.githubusercontent.com/MattWillFlood/EntropyHub/main/ExampleData/" + SigName + ".txt"
    resp = requests.get(url)    
    
    if SigName == 'chirp':
        Temp = np.matrix([x for x in resp.text.split('\n')][2:-1])    
        X = np.squeeze(np.array(Temp).astype(float))
    
    elif SigName in ['uniform','gaussian','randintegers']:    
        Temp = np.matrix([x for x in resp.text.split('\n')][2])    
        X = np.squeeze(np.array(Temp))
    
    else: 
        Temp = ([np.matrix([y for y in x.split()])
                         for x in resp.text.split('\r\n')[2:-1]])                         
        X = np.squeeze(np.array(Temp).astype(float))
            
    return X

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Cosine Similarity Entropy function."""
import numpy as np 

def CoSiEn(Sig,  m=2, tau=1, r=0.1, Logx=2, Norm=0):
    """CoSiEn  estimates the cosine similarity entropy of a univariate data sequence.
    
    .. code-block:: python
        
        CoSi, Bm = CoSiEn(Sig) 
        
    Returns the cosine similarity entropy (``CoSi``) and the corresponding
    global probabilities (``Bm``) estimated from the data sequence (``Sig``) using the
    default parameters:   embedding dimension = 2, time delay = 1, 
    angular threshold = .1,  logarithm = base 2,

    .. code-block:: python
        
        CoSi, Bm = CoSiEn(Sig, keyword = value, ...)
        
    Returns the cosine similarity entropy (``CoSi``) estimated from the data
    sequence (``Sig``) using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, an integer > 1
        :tau:   - Time Delay, a positive integer
        :r:     - Angular threshold, a value in range [0 < ``r`` < 1]
        :Logx:  - Logarithm base, a positive scalar (enter 0 for natural log) 
        :Norm:  - Normalisation of ``Sig``, one of the following integers:
            
                  0.  no normalisation - default
                  1.  normalises ``Sig`` by removing median(``Sig``)
                  2.  normalises ``Sig`` by removing mean(``Sig``)
                  3.  normalises ``Sig`` w.r.t. SD(``Sig``)
                  4.  normalises ``Sig`` values to range [-1 1]

    :See also:
        ``PhasEn``, ``SlopEn``, ``GridEn``, ``MSEn``, ``hMSEn``
    
    :References:
        [1] Theerasak Chanwimalueang and Danilo Mandic,
            "Cosine similarity entropy: Self-correlation-based complexity
            analysis of dynamical systems."
            Entropy 
            19.12 (2017): 652.
            
    """
    
    Sig = np.squeeze(Sig)
    N = Sig.shape[0]  
    if Logx == 0:
        Logx = np.exp(1)
        
    assert N>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 1), "m:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r, (int,float)) and r >= 0 and r <= 1, "r:    must be a value in range [0 1]"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    assert isinstance(Norm,int) and np.isin(Norm, np.arange(5)), "Norm:     must be an integer in range [0 4]"
                   
    if Norm == 1:
        Xi = Sig - np.median(Sig)
    elif Norm == 2:
        Xi = Sig - np.mean(Sig)
    elif Norm == 3:
        Xi = (Sig - np.mean(Sig))/np.std(Sig)
    elif Norm == 4:
        Xi = (2*(Sig - min(Sig))/np.ptp(Sig)) - 1
    else:
        Xi = Sig        
    
    Nx = N - (m-1)*tau
    Zm = np.zeros((Nx,m))
    for n in range(m):
        Zm[:,n] = Xi[n*tau:Nx+(n*tau)]
        
    Num = np.inner(Zm,Zm)
    #Mag = np.linalg.norm(Zm, axis=1) # Dont know which is faster...
    Mag = np.sqrt(np.diag(Num))
    Den = np.outer(Mag, Mag)
                
    AngDis = np.arccos(np.triu(Num/Den,1))/np.pi    
    if np.max(np.imag(AngDis)) < 10**-6:
        Bm = np.sum(np.triu(np.round(AngDis,6) < r,1))/(Nx*(Nx-1)/2)    
    else:
        Bm = np.sum(np.triu(np.real(AngDis) < r,1))/(Nx*(Nx-1)/2)
        print('Warning: Complex values ignored')
    if Bm == 1 or Bm == 0:
        CoSi = np.NaN
    else:
        CoSi = -(Bm*np.log(Bm)/np.log(Logx)) - ((1-Bm)*np.log(1-Bm)/np.log(Logx))

    return CoSi, Bm

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""WindowData restructures a univariate/multivariate dataset into a collection of subsequence windows."""
import numpy as np 

def WindowData(Data, WinLen=None, Overlap=0, Mode="exclude"):
    """WindowData restructures a univariate/multivariate dataset into a collection of subsequence windows.

    .. code-block:: python

       WinData, Log = WindowData(Data) 
 
   Windows the sequence(s) given in ``Data`` into a collection of subsequnces  
   of floor(N/5) elements with no overlap, excluding any remainder
   elements that do not fill the final window.
   If ``Data`` is a univariate sequence (vector), ``Windata`` is a tuple of 5
   vectors. If ``Data`` is a set of multivariate sequences (NxM matrix), 
   each of M columns is treated as a sequence with N elements 
   and ``WinData`` is a tuple of 5 matrices of size [(floor*N,5), M]. 
   The ``Log`` dictionary contains information about the windowing process, including:
       :DataType:      - The type of data sequence passed as ``Data``
       :DataLength:    - The number of sequence elements in ``Data``
       :WindowLength:  - The number of elements in each window of ``WinData``
       :WindowOverlap: - The number of overlapping elements between windows
       :TotalWindows:  - The number of windows extracted from ``Data`` 
       :Mode:          - Decision to include or exclude any remaining sequence elements (``< WinLen``) that do not fill the window.
 
    .. code-block:: python

       WinData, Log = WindowData(Data, keyword = value, ...)
 
   Windows the sequence(s) given in ``Data`` into a collection of subsequnces
   using the specified keyword arguments: 
       :WinLen:  - Number of elements in each window, a positive integer (>10)
       :Overlap: - Number of overlapping elements between windows, a positive integer (< WinLen)
       :Mode:    - Decision to include or exclude any remaining sequence elements (< ``WinLen``) that do not fill the window, a string - either ``"include"`` or ``"exclude"`` (default).

   :See also:
       ``ExampleData``

    """
               
    Data = np.squeeze(Data)
    assert isinstance(Data,np.ndarray) and Data.shape[0]>10 and Data.ndim<=2,  "Data:   must be a numpy Vector (length N) or an NxM numpy matrix where N>10 and M>1"

    if Data.ndim == 1:   
        DataType =  "single univariate vector (1 sequence)"
        N = Data.shape[0]; Dn = 0 
    elif Data.ndim == 2: 
        N, Dn = Data.shape 
        DataType = "multivariate matrix (" +str(Dn)+" vectors)"

    if WinLen is None: WinLen = N//5
    assert isinstance(WinLen,int)  and WinLen<N and WinLen>10, "WinLen: must be an integer such that 10 < WinLen < N"
    assert isinstance(Overlap,int) and Overlap>=0 and Overlap<WinLen, "Overlap: The number of overlapping window samples such that 0 <= Overlap < WinLen"
    assert isinstance(Mode, str) and  Mode.lower() in ["exclude","include"], "Mode: Option to include/exclude samples that do not fill final window, either 'exclude' or 'include'"
        
    if Dn==0:  Data = np.expand_dims(Data, 1)
    M = (N - Overlap)//(WinLen - Overlap)
    Step = WinLen-Overlap
    Xout = []
    for k in range(0,M*Step,Step):
        Xout.append(np.squeeze(Data[k:k+WinLen,:]))   
    
    if (Mode.lower() == "include") and (len(Xout)-1)*Step+WinLen!=N:   
        Xout.append(np.squeeze(Data[k+Step:,:])); M+=1
    
    WinData = tuple(Xout)
    Log = {"DataType": DataType, "DataLength": N, "WindowLength": WinLen,
           "WindowOverlap": Overlap,  "TotalWindows": M,   "Mode": Mode}
    
    return WinData, Log


"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Cross Approximate Entropy function."""
import numpy as np    

def XApEn(*Sig, m=2, tau=1, r=None, Logx=np.exp(1)):
    """XApEn  estimates the cross-approximate entropy between two univariate data sequences.

    .. code-block:: python
    
        XAp, Phi = XApEn(Sig1, Sig2)
        
    Returns the cross-approximate entropy estimates (``XAp``) and the average
    number of matched vectors (``Phi``) for ``m`` = [0,1,2], estimated for the data
    sequences contained in ``Sig1`` and ``Sig2`` using the default parameters:
    embedding dimension = 2, time delay = 1, 
    radius threshold = 0.2*SDpooled(``Sig1``,``Sig2``),  logarithm = natural
    
    **NOTE: ``XApEn`` is direction-dependent. Thus, ``Sig1`` is used as the template data sequence, and ``Sig2`` is the matching sequence.
 
     .. code-block:: python
       
         XAp, Phi = XApEn(Sig1, Sig2, keyword = value, ...)
        
    Returns the cross-approximate entropy estimates (``XAp``) between the data
    sequences contained in ``Sig1`` and ``Sig2`` using the specified 'keyword' arguments:
        :m:     - Embedding Dimension,  a positive integer   [default: 2]
        :tau:   - Time Delay, a positive integer        [default: 1]
        :r:     - Radius Distance Threshold, a positive scalar [default: 0.2*SDpooled(``Sig1``,``Sig2``)]
        :Logx:  - Logarithm base, a positive scalar     [default: natural]
 
    :See also:
        ``XSampEn``, ``XFuzzEn``, ``XMSEn``, ``ApEn``, ``SampEn``, ``MSEn``
    
    :References:
        [1] Steven Pincus and Burton H. Singer,
            "Randomness and degrees of irregularity." 
            Proceedings of the National Academy of Sciences 
            93.5 (1996): 2083-2088.
    
        [2] Steven Pincus,
            "Assessing serial irregularity and its implications for health."
            Annals of the New York Academy of Sciences 
            954.1 (2001): 245-267.
    
    """

    assert len(Sig)<=2,  """Input arguments to be passed as data sequences:
        - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
        - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
    if len(Sig)==1:
        Sig = np.squeeze(Sig)
        assert max(Sig.shape)>10 and min(Sig.shape)==2,  """Input arguments to be passed as data sequences:
            - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
            - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
        if Sig.shape[0] == 2:
            Sig = Sig.transpose()  
        S1 = Sig[:,0]; S2 = Sig[:,1]     
        
    elif len(Sig)==2:
        S1 = np.squeeze(Sig[0])
        S2 = np.squeeze(Sig[1])
        
    N  = S1.shape[0]
    N2 = S2.shape[0]
    if r is None:
        r = 0.2*np.sqrt((np.var(S1)*(N-1) + np.var(S2)*(N2-1))/(N+N2-1))         
         
    assert N>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"
    assert isinstance(m,int) and (m > 0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) and (r>=0), "r:     must be a positive value"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    
    Counter = 1*(abs(np.expand_dims(S1,axis=1) - np.expand_dims(S2,axis=0))<= r)
    M = np.hstack((m*np.ones(N-m*tau), np.repeat(np.arange(m-1,0,-1),tau)))
    XAp = np.zeros((m+1))
    Phi = np.zeros((m+2))    
    for n in range(M.shape[0]):
        ix = np.where(Counter[n, :] == 1)[0]        
        for k in range(1, int(M[n]+1)):
            ix = ix[ix + (k*tau) < N2]     
            if not len(ix):   break            
            p1 = np.tile(S1[n:n+1+(tau*k):tau], (ix.shape[0], 1))                       
            p2 = S2[np.expand_dims(ix,axis=1) + np.arange(0,(k*tau)+1,tau)]           
            ix = ix[np.amax(abs(p1-p2), axis=1) <= r] 
            Counter[n, ix] += 1
            
    #Phi[0] = (np.log(N)/np.log(Logx))/N
    # Phi[1] = np.mean(np.log(np.sum(Counter>0,axis=0)/N)/np.log(Logx))
    Temp = np.sum(Counter>0,axis=0); Temp = Temp[Temp!=0]
    Phi[1] = np.mean(np.log(Temp/N)/np.log(Logx))
    XAp[0]  = Phi[0] - Phi[1]
    
    for k in range(m):
        ai = np.sum(Counter>k+1,axis=0)/(N-(k+1)*tau)
        bi = np.sum(Counter>k,axis=0)/(N-(k*tau))
        ai = ai[ai>0]
        bi = bi[bi>0]
        
        with np.errstate(divide='ignore', invalid='ignore'):
            Phi[k+2] = np.sum(np.log(ai)/np.log(Logx))/(N-(k+1)*tau)
            XAp[k+1]  = np.sum(np.log(bi)/np.log(Logx))/(N-(k*tau)) - Phi[k+2]
        
    return XAp, Phi

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""




from EntropyHub._ApEn import ApEn
from EntropyHub._AttnEn import AttnEn
from EntropyHub._BubbEn import BubbEn
from EntropyHub._CondEn import CondEn
from EntropyHub._CoSiEn import CoSiEn
from EntropyHub._DistEn import DistEn
from EntropyHub._DispEn import DispEn
from EntropyHub._DivEn import DivEn
from EntropyHub._EnofEn import EnofEn
from EntropyHub._FuzzEn import FuzzEn
from EntropyHub._GridEn import GridEn
from EntropyHub._IncrEn import IncrEn
from EntropyHub._K2En import K2En
from EntropyHub._PhasEn import PhasEn
from EntropyHub._PermEn import PermEn
from EntropyHub._RangEn import RangEn
from EntropyHub._SampEn import SampEn
from EntropyHub._SpecEn import SpecEn
from EntropyHub._SlopEn import SlopEn
from EntropyHub._SyDyEn import SyDyEn


from EntropyHub._XApEn import XApEn
from EntropyHub._XCondEn import XCondEn 
from EntropyHub._XDistEn import XDistEn 
from EntropyHub._XFuzzEn import XFuzzEn
from EntropyHub._XK2En import XK2En
from EntropyHub._XPermEn import XPermEn
from EntropyHub._XSampEn import XSampEn
from EntropyHub._XSpecEn import XSpecEn


from EntropyHub._SampEn2D import SampEn2D
from EntropyHub._FuzzEn2D import FuzzEn2D
from EntropyHub._DistEn2D import DistEn2D
from EntropyHub._DispEn2D import DispEn2D
from EntropyHub._EspEn2D import EspEn2D
from EntropyHub._PermEn2D import PermEn2D


from EntropyHub._MvSampEn import MvSampEn
from EntropyHub._MvFuzzEn import MvFuzzEn
from EntropyHub._MvCoSiEn import MvCoSiEn
from EntropyHub._MvDispEn import MvDispEn
from EntropyHub._MvPermEn import MvPermEn


from EntropyHub._ExampleData import ExampleData
from EntropyHub._WindowData import WindowData
from EntropyHub._MSobject import MSobject


from EntropyHub._MSEn import MSEn
from EntropyHub._cMSEn import cMSEn
from EntropyHub._rMSEn import rMSEn
from EntropyHub._hMSEn import hMSEn

from EntropyHub._XMSEn import XMSEn
from EntropyHub._cXMSEn import cXMSEn
from EntropyHub._rXMSEn import rXMSEn
from EntropyHub._hXMSEn import hXMSEn

from EntropyHub._MvMSEn import MvMSEn
from EntropyHub._cMvMSEn import cMvMSEn

# from EntropyHub._MSEn2D import MSEn2D


def greet():
    print("""
    
    	 ___  _   _  _____  _____  ____  ____  _     _          
    	|  _|| \ | ||_   _||     \|    ||    || \   / |   ___________ 
    	| \_ |  \| |  | |  |   __/|    ||  __| \ \_/ /   /  _______  \\
    	|  _|| \ \ |  | |  |   \  |    || |     \   /   |  /  ___  \  |
    	| \_ | |\  |  | |  | |\ \ |    || |      | |    | |  /   \  | | 
    	|___||_| \_|  |_|  |_| \_||____||_|      |_|   _|_|__\___/  | | 
    	 _   _  _   _  ____                           / |__\______\/  | 
    	| | | || | | ||    \     An open-source      |  /\______\__|_/ 
    	| |_| || | | ||    |     toolkit for         | |  /   \  | | 
    	|  _  || | | ||    \     entropic time-      | |  \___/  | |          
    	| | | || |_| ||     \    series analysis     |  \_______/  |
    	|_| |_|\_____/|_____/                         \___________/ 
    
          
        Please use the following citation on any scientific outputs achieved with the help of EntropyHub:
        
        Matthew W. Flood,
        EntropyHub: An Open-Source Toolkit for Entropic Time Series Analysis,
        PLoS One 16(11):e0259448 (2021),
        DOI: 10.1371/journal.pone.0259448

        www.EntropyHub.xyz
        
        """)
    
def functionlist():
    print("""    
        EntropyHub functions belong to one of five main classes/categories:
        Base Entropies             >>  e.g. Approximate Entropy (ApEn),
                                            Sample Entropy (SampEn)
        Cross Entropies            >>  e.g. Cross-Approximate Entropy (XApEn)
                                            Cross-Sample Entropy (XSampEn)
        Bidimensional Entropies    >>  e.g. Bidimensional Sample Entropy (SampEn2D)
                                            Bidimensional Fuzzy Entropy (FuzzEn2D)
        Multiscale Entropies       >>  e.g. Multiscale Sample Entropy (MSEn)
                                            Refined Multiscale Sample Entropy (rMSEn)
                                            Composite Multiscale Sample Entropy (cMSEn)
        Multiscale Cross Entropies >>  e.g. Multiscale Cross-Sample Entropy (XMSEn)
                                            Refined Multiscale Cross-Sample Entropy (rXMSEn)
        
        _________________________________________________________________________
        Base Entropies                                     	|	Function Name
        ____________________________________________________|__________________
        Approximate Entropy                               	|	ApEn
        Sample Entropy                                    	|	SampEn
        Fuzzy Entropy                                     	|	FuzzEn
        Kolmogorov Entropy                                	|	K2En
        Permutation Entropy                               	|	PermEn
        Conditional Entropy                               	|	CondEn
        Distribution Entropy                              	|	DistEn
        Spectral Entropy                                  	|	SpecEn
        Dispersion Entropy                                	|	DispEn
        Symbolic Dynamic Entropy                          	|	SyDyEn
        Increment Entropy                                 	|	IncrEn
        Cosine Similarity Entropy                         	|	CoSiEn
        Phase Entropy                                      	|	PhasEn
        Slope Entropy                                      	|	SlopEn
        Bubble Entropy                                      |	BubbEn
        Gridded Distribution Entropy                        |	GridEn
        Entropy of Entropy                                  |	EnofEn
        Attention Entropy                                   |	AttnEn
        Diversity Entropy                                   |	DivEn
        Range Entropy                                       |	RangEn
        
        _______________________________________________________________________
        Cross Entropies                                     |	Function Name
        ____________________________________________________|__________________
        Cross Sample Entropy                                |	XSampEn
        Cross Approximate Entropy                           |	XApEn
        Cross Fuzzy Entropy                                 |	XFuzzEn
        Cross Permutation Entropy                           |	XPermEn
        Cross Conditional Entropy                           |	XCondEn
        Cross Distribution Entropy                          |	XDistEn
        Cross Spectral Entropy                              |	XSpecEn
        Cross Kolmogorov Entropy                           	|	XK2En
        	
        _________________________________________________________________________
        Bidimensional Entropies                              |	Function Name
        _____________________________________________________|___________________
        Bidimensional Sample Entropy                         |	SampEn2D
        Bidimensional Fuzzy Entropy                          |	FuzzEn2D
        Bidimensional Distribution Entropy                   |	DistEn2D
        Bidimensional Dispersion Entropy                     |	DispEn2D
        Bidimensional Permutation Entropy                    |	PermEn2D
        Bidimensional Espinosa Entropy                       |	EspEn2D
        
        _________________________________________________________________________
        Multivariate Entropies                               |	Function Name
        _____________________________________________________|___________________
        Multivariate Sample Entropy                          |	MvSampEn
        Multivariate Fuzzy Entropy                           |	MvFuzzEn
        Multivariate Cosine Similarity Entropy               |	MvCoSiEn
        Multivariate Dispersion Entropy                      |	MvDispEn
        Multivariate Permutation Entropy                     |	MvPermEn
        	
        _________________________________________________________________________
        Multiscale Entropy Functions                          | Function Name
        ______________________________________________________|__________________ 
        Multiscale Entropy Object                             |   MSobject
                                                              |
        Multiscale Entropy                                    |   MSEn
        Composite/Refined-Composite Multiscale Entropy        |   cMSEn
        Refined Multiscale Entropy                            |   rMSEn
        Hierarchical Multiscale Entropy Object                |   hMSEn
        
        _________________________________________________________________________
        Multiscale Entropies	MSEn                          |	Function Name
        ______________________________________________________|__________________
        Multiscale Sample Entropy                             |	
        Multiscale Approximate Entropy                        |	
        Multiscale Fuzzy Entropy                              |	
        Multiscale Permutation Entropy                        |	
        Multiscale Dispersion Entropy                         |	
        Multiscale Cosine Similarity Entropy                  |	
        Multiscale Symblic Dynamic Entropy                    |	MSobject
        Multiscale Conditional Entropy                        |	     +
        Multiscale Entropy of Entropy                         |   MSEn / cMSEn
        Multiscale Gridded Distribution Entropy               |	rMSEn / hMSEn
        Multiscale Slope Entropy                              |
        Multiscale Phase Entropy                              |		
        Multiscale Kolmogorov Entropy                         |	
        Multiscale Distribution Entropy                       |	
        Multiscale Bubble Entropy                             |	
        Multiscale Increment Entropy                          |	
        Multiscale Attention Entropy                          |	
        Multiscale Diversity Entropy                          |	
        Multiscale Range Entropy                              |	
        	
        _________________________________________________________________________
        Multiscale Cross-Entropy Functions                    |   Function Name
        ______________________________________________________|__________________
        Multiscale Cross-Entropy Object                       |   MSobject
                                                              |
        Multiscale Cross-Entropy                              |   XMSEn
        Composite/Refined-Composite Multiscale Cross-Entropy  |   cXMSEn
        Refined Multiscale Entropy                            |   rXMSEn
        Hierarchical Multiscale Entropy Object                |   hXMSEn
        
        _________________________________________________________________________
        Multiscale Cross-Entropies                            |	Function Name
        ______________________________________________________|__________________
        Multiscale Cross-Sample Entropy                       |	
        Multiscale Cross-Approximate Entropy                  |	
        Multiscale Cross-Fuzzy Entropy                        |	MSobject
        Multiscale Cross-Permutation Entropy                  |	    +
        Multiscale Cross-Distribution Entropy                 |	XMSEn / cXMSEn
        Multiscale Cross-Kolmogorov Entropy                   |   rXMSEn / hXMSEn
        Multiscale Cross-Conditional Entropy                  |	
        
        
        _________________________________________________________________________
        Multivariate Multiscale Entropy Functions             |	Function Name
        ______________________________________________________|__________________
        Multiscale Cross-Entropy Object                       |   MSobject
                                                              |
        Multivariate Multiscale Entropy                       |	 MvMSEn
        Composite/Refined-Composite Multiscale Cross-Entropy  |	 cMvMSEn
        
        
        _________________________________________________________________________
        Multivariate Multiscale Entropies                     |	Function Name
        ______________________________________________________|__________________
        Multivariate Multiscale Sample Entropy                |	
        Multivariate Multiscale Fuzzy Entropy                 |	MSobject
        Multivariate Multiscale Dispersion Entropy            |	    +
        Multivariate Multiscale Permutation Entropy           |	MvMSEn / cMvMSEn
        Multivariate Multiscale Cosine Similarity Entropy     |	
        
        
        _________________________________________________________________________
        Other  Functions                                      |	Function Name
        ______________________________________________________|__________________
        Example Dataset Import Tool                           |	 ExampleData
        Data Windowing Tool                                   |	 WindowData
        
        
        
        This package is open for use by all in accordance with the terms of the 
        attached License agreement. Any scientific outputs obtained using 
        EntropyHub are required to include the following citation:
        
          Matthew W. Flood, 
          "EntropyHub: An open-source toolkit for entropic time series analysis",
          2021, https://github.com/MattWillFlood/EntropyHub
        
          © Copyright 2024 Matthew W. Flood, EntropyHub
        
          Licensed under the Apache License, Version 2.0 (the "License");
          you may not use this file except in compliance with the License.
          You may obtain a copy of the License at
        
               http://www.apache.org/licenses/LICENSE-2.0
        
          Unless required by applicable law or agreed to in writing, software
          distributed under the License is distributed on an "AS IS" BASIS,
          WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
          See the License for the specific language governing permissions and
          limitations under the License.
        
          For Terms of Use see https://github.com/MattWillFlood/EntropyHub
        
        """)


        # _________________________________________________________________________
        # Multiscale Bidimensional Entropy Functions            |   Function Name
        # ______________________________________________________|__________________
        # Multiscale Bidimensional Entropy Object               |   MSobject
        #                                                       |
        # Multiscale Bidimensional Entropy                      |   MSEn2D
        
        # _________________________________________________________________________
        # Multiscale Bidimensional-Entropies                    |	Function Name
        # ______________________________________________________|__________________
        # Multiscale Bidimensional Sample Entropy               |	
        # Multiscale Bidimensional Fuzzy Entropy                |	
        # Multiscale Bidimensional Distribution Entropy         |	MSobject
        # Multiscale Bidimensional Dispersion Entropy           |	    +
        # Multiscale Bidimensional Permutation Entropy          |	 MSEn2D
        # Multiscale Bidimensional Espinosa Entropy             |



"Base refined Multiscale Entropy function."
import numpy as np   
from scipy.signal import butter, filtfilt
from matplotlib.pyplot import figure, axes, show
from copy import deepcopy

def rMSEn(Sig, Mbjx, Scales=3, F_Order=6, F_Num=.5, RadNew=0, Plotx=False):
    """rMSEn  returns the refined multiscale entropy of a univariate data sequence.

    .. code-block::
        
        MSx, CI = rMSEn(Sig, Mobj) 
        
    Returns a vector of refined multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data sequence (``Sig``) using the parameters specified by
    the multiscale object (``Mobj``) and the following default parameters:
    Scales = 3, Butterworth LPF Order = 6, Butterworth LPF cutoff frequency
    at scale (T): Fc = 0.5/T. 
    If the entropy function specified by ``Mobj`` is ``SampEn`` or ``ApEn``, ``rMSEn``
    updates the threshold radius of the data sequence (Xt) at each scale
    to 0.2*std(Xt) if no ``r`` value is provided by Mobj, or r*std(Xt) if ``r``
    is specified.
     
    .. code-block:: python
    
        MSx, CI = rMSEn(Sig, Mobj, keyword = value, ...)
        
    Returns a vector of refined multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data sequence (``Sig``) using the parameters specified by
    the multiscale object (``Mobj``) and the following 'keyword' arguments:
        :Scales:   - Number of temporal scales, a positive integer   (default: 3)
        :F_Order:  - Butterworth low-pass filter order, a positive integer (default: 6)
        :F_Num:    - Numerator of Butterworth low-pass filter cutoff frequency, a scalar value in range [0 < ``F_Num`` < 1].  The cutoff frequency  at each scale (T) becomes: Fc = F_Num/T.  (default: 0.5)
        :RadNew:   - Radius rescaling method, an integer in the range [1 4].
                     When the entropy specified by ``Mobj`` is ``SampEn`` or ``ApEn``, RadNew rescales the radius threshold in each sub-sequence
                     at each time scale (Ykj). If a radius value (``r``) is specified by ``Mobj``, this becomes the rescaling coefficient, otherwise
                     it is set to 0.2 (default). The value of RadNew specifies one of the following methods:
                         
                        * [1] Standard Deviation          - ``r*std(Ykj)``
                        * [2] Variance                    - ``r*var(Ykj)``
                        * [3] Mean Absolute Deviation     - ``r*mad(Ykj)``
                        * [4] Median Absolute Deviation   - ``r*mad(Ykj,1)``
                     
        :Plotx:    - When ``Plotx == True``, returns a plot of the entropy value at each time scale (i.e. the multiscale entropy curve) [default: False]
    
    :See also:
        ``MSobject``, ``MSEn``, ``cMSEn``, ``hMSEn``, ``XMSEn``, ``rXMSEn``, ``SampEn``, ``ApEn``
    
    :References:
        [1] Madalena Costa, Ary Goldberger, and C-K. Peng,
            "Multiscale entropy analysis of complex physiologic time series."
            Physical review letters
            89.6 (2002): 068102.
    
        [2] Vadim V. Nikulin, and Tom Brismar,
            "Comment on “Multiscale entropy analysis of complex physiologic
            time series”." 
            Physical review letters 
            92.8 (2004): 089803.
    
        [3] Madalena Costa, Ary L. Goldberger, and C-K. Peng. 
            "Costa, Goldberger, and Peng reply." 
            Physical Review Letters
            92.8 (2004): 089804.
    
        [4] José Fernando Valencia, et al.,
            "Refined multiscale entropy: Application to 24-h holter 
            recordings of heart period variability in healthy and aortic 
            stenosis subjects." 
            IEEE Transactions on Biomedical Engineering 
            56.9 (2009): 2202-2213.
    
        [5] Puneeta Marwaha and Ramesh Kumar Sunkaria,
            "Optimal selection of threshold value ‘r’for refined multiscale
            entropy." 
            Cardiovascular engineering and technology 
            6.4 (2015): 557-576.
    
    """

    Mobj = deepcopy(Mbjx)    
    Sig = np.squeeze(Sig)    
    assert Sig.shape[0]>10 and Sig.ndim == 1, "Sig:   must be a numpy vector"    
    assert isinstance(Mobj,object), "Mobj:  must be a multiscale entropy \
    object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert isinstance(F_Order, int) and F_Order>0, \
    "F_Order:    a positive integer"
    assert isinstance(F_Num,float) and 0<F_Num<1, \
    "F_Num:     a scalar value in range [0 < F_Num < 1]"
    assert (np.isin(RadNew,np.arange(1,5)) and Mobj.Func.__name__ in \
    ['SampEn','ApEn']) or RadNew==0, "RadNew:     must be an integer in range \
    [1 4] and entropy function must be 'SampEn' or 'ApEn'"
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
    assert Mobj.Func.__name__[0].lower() != 'x', ("Base entropy estimator is a cross-entropy method." 
    "To perform refined multiscale CROSS-entropy estimation, use rXMSEn.")
    
    MSx = np.zeros(Scales)    
    if (Mobj.Func.__name__ in ['SampEn','ApEn']) and RadNew==0:
        RadNew = 1
    
    if Mobj.Func.__name__ in 'SampEn': Mobj.Kwargs['Vcp'] = False 
    
    if RadNew > 0:
        if RadNew == 1:
            Rnew = lambda x: np.std(x)
        elif RadNew == 2:
            Rnew = lambda x: np.var(x)
        elif RadNew == 3:
            Rnew = lambda x: np.mean(abs(x-np.mean(x)))
        elif RadNew == 4:
            Rnew = lambda x: np.median(abs(x-np.median(x)))    
        
        try:
            Cx = Mobj.Kwargs.get('r')*1
        except:
            Cy = ['Standard Deviation','Variance','Mean Abs Deviation',
                  'Median Abs Deviation']
            print('WARNING: No radius value provided.\nDefault set to ' \
                  '0.2*(%s) of each new time-series.'%Cy[RadNew-1])            
            Cx = .2

    for T in range(1,Scales+1):
        print(' .', end='')
        Temp = refined(Sig,T,F_Order,F_Num)   
        if RadNew > 0:
            Mobj.Kwargs.update({'r': Cx*Rnew(Temp)})             
        Tempx = Mobj.Func(Temp,**Mobj.Kwargs)        
        
        if isinstance(Tempx,tuple):
            if isinstance(Tempx[0],(int,float)):
                MSx[T-1] = Tempx[0]
            else:
                MSx[T-1] =Tempx[0][-1]
        elif isinstance(Tempx,(int,float)):
                MSx[T-1] = Tempx
        elif isinstance(Tempx,np.ndarray):
                MSx[T-1] = Tempx[-1]
                    
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')

    if Plotx:
       figure()
       ax1 = axes()   
       ax1.plot(np.arange(1,Scales+1), MSx, color=(8/255, 63/255, 77/255), linewidth=3)
       ax1.scatter(np.arange(1,Scales+1), MSx, 60, color=(1,0,1))
       ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_title('Refined Multiscale %s'%(Mobj.Func.__name__), 
                     fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))      
       show()
    
    return MSx, CI

def refined(Z,sx,P1,P2):
    bb, aa = butter(P1, P2/sx)
    Yt = filtfilt(bb, aa, Z)
    Y = Yt[::sx]
    return Y


"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"Base Multiscale Bidimensional Entropy function."
import numpy as np   
from matplotlib.pyplot import figure, axes, show
from copy import deepcopy
from scipy.signal import convolve2d 

def MSEn2D(Mat, Mbjx, Scales=3, Methodx='coarse', RadNew=0, Plotx=False):
    """MSEn2D  Returns the multiscale entropy of a bidimensional data matrix.
    
    .. code-block:: python
    
        MSx,CI = MSEn(Mat, Mobj) 
    
    Returns a vector of multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data matrix ``Mat`` using the parameters specified 
    by the multiscale object (``Mobj``) over 3 temporal scales with coarse-
    graining (default). 
  
    .. code-block:: python
    
        MSx,CI = MSEn(Mat, Mobj, keyword = value, ...)
        
    Returns a vector of multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data matrix ``Mat`` using the parameters specified by
    the multiscale object (``Mobj``) and the following ``keyword`` arguments:
        :Scales:   - Number of temporal scales, an integer > 1   (default: 3)
        :Methodx:  - Graining method, one of the following: [default: ``'coarse'``]  {``'coarse'``, ``'modified'``}
        :RadNew:   - Radius rescaling method, an integer in the range [1 4].
                     When the entropy specified by ``Mobj`` is ``SampEn2D``, RadNew rescales the radius threshold in each sub-matrix
                     at each time scale (Ykj). If a radius value (``r``) is specified by ``Mobj``, this becomes the rescaling coefficient, otherwise
                     it is set to 0.2 (default). The value of RadNew specifies one of the following methods:
                         
                        * [1] Standard Deviation          - ``r*std(Ykj)``
                        * [2] Variance                    - ``r*var(Ykj)``
                        * [3] Mean Absolute Deviation     - ``r*mad(Ykj)``
                        * [4] Median Absolute Deviation   - ``r*mad(Ykj,1)``
                         
        :Plotx:    - When ``Plotx == True``, returns a plot of the entropy value at each time scale (i.e. the multiscale entropy curve) [default: False]
    
    For further info on these graining procedures see the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_.
    
    :See also:
        ``MSobject``, ``SampEn2D``, ``DispEn2D``, ``FuzzEn2D``, ``DistEn2D``, ``PermEn2D``, ``EspEn2D``
    
    :References:
         [1] Luiz E.V. Silva, et al.,
              "Two-dimensional multiscale entropy analysis: Applications to image texture evaluation."
              Signal Processing
              147 (2018): 224-232.
    
         [2] Cristina Morel and Anne Humeau-Heurtier
              "Multiscale permutation entropy for two-dimensional patterns." 
              Pattern Recognition Letters 
              150 (2021): 139-146.
    
         [3] Mirvana Hilal and Anne Humeau-Heurtier
              "Bidimensional Fuzzy Entropy: Principle Analysis and Biomedical Applications"
              IEEE Engineering in Medicine and Biology Society (EMBS) Conference
              2019: 4811-4814
    
    """
    
    Mobj = deepcopy(Mbjx)
    
    Mat = np.squeeze(Mat)    
    Chk = ['coarse','modified']    
    assert Mat.ndim==2 and min(Mat.shape)>10 , "Mat:   must be a 2D numpy array with height & width > 10"
    assert isinstance(Mobj,object), "Mobj:  must be a multiscale entropy \
    object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert Methodx.lower() in Chk, "Methodx:  must be one of the following string names - 'coarse', 'modified'"    
    assert isinstance(RadNew,int) and (np.isin(RadNew,np.arange(1,5)) \
                and Mobj.Func.__name__ in ['SampEn2D']) or RadNew==0, \
    "RadNew:     must be 0, or an integer in range [1 4] with entropy function 'SampEn2D'"
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
        
    Func2 = globals()[Methodx.lower()]
    MSx = np.zeros(Scales)
    
    if RadNew > 0:
        if RadNew == 1:
            Rnew = lambda x: np.std(x)
        elif RadNew == 2:
            Rnew = lambda x: np.var(x)
        elif RadNew == 3:
            Rnew = lambda x: np.mean(abs(x-np.mean(x)))
        elif RadNew == 4:
            Rnew = lambda x: np.median(abs(x-np.median(x)))    
        
        try:
            Cx = Mobj.Kwargs.get('r')*1
        except:
            Cy = ['Standard Deviation','Variance','Mean Abs Deviation',
                  'Median Abs Deviation']
            print('WARNING: No radius value provided.\nDefault set to ' \
                  '0.2*(%s) of each new time-series.'%Cy[RadNew-1])            
            Cx = .2

    for T in range(1,Scales+1):
        print(' .', end='')
        Temp = Func2(Mat,T)        
        
        if RadNew > 0:
            Mobj.Kwargs.update({'r': Cx*Rnew(Temp)})            
        Tempx = Mobj.Func(Temp,**Mobj.Kwargs)
                    
        if isinstance(Tempx,tuple):
           Temp2 = Tempx[0]  
        else:
           Temp2 = Tempx
                    
        MSx[T-1] = Temp2
                
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')

    if Plotx:
       figure()
       ax1 = axes()   
       ax1.plot(np.arange(1,Scales+1), MSx, color=(8/255, 63/255, 77/255), linewidth=3)
       ax1.scatter(np.arange(1,Scales+1), MSx, 60, color=(1,0,1))
       ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
       ax1.set_title('Multiscale %s (%s-graining method)'%(Mobj.Func.__name__,Methodx), 
                     fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))      
       show()
    
    return MSx, CI


def coarse(Z,sx):   
    Nh = Z.shape[0]//sx
    Nw = Z.shape[1]//sx    
    Y = np.mean(np.reshape(Z[:sx*Nh,:sx*Nw],(sx,Nh,sx,Nw)),axis=(0,2))
    return Y

def modified(Z,sx):
    Y = convolve2d(Z, np.ones((sx,sx)), mode='valid')/(sx*sx)
    return Y 

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Cross Permutation Entropy function."""
import numpy as np

def XPermEn(*Sig, m=3, tau=1, Logx=np.exp(1)):
    """XPermEn  estimates the cross-permutation entropy between two univariate data sequences.

    .. code-block:: python
    
        XPerm = XPermEn(Sig1, Sig2) 
        
    Returns the cross-permuation entropy estimates (``XPerm``) estimated betweeen
    the data sequences contained in ``Sig1`` and ``Sig2`` using the default parameters:
    embedding dimension = 3, time delay = 1, logarithm = base 2, 
    
    .. code-block:: python
 
        XPerm = XPermEn(Sig1, Sig2, keyword = value, ...)
        
    Returns the permutation entropy estimates (``Perm``) estimated between the data
    sequences contained in ``Sig1`` and ``Sig2`` using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, an integer > 2   [default: 3]
            **Note: ``XPerm`` is undefined for embedding dimensions < 3.
        :tau:   - Time Delay, a positive integer        [default: 1]
        :Logx:  - Logarithm base, a positive scalar     [default: 2]   (enter 0 for natural log). 
 
    :See also:
        ``PermEn``, ``XApEn``, ``XSampEn``, ``XFuzzEn``, ``XMSEn``
  
    :References:
        [1] Wenbin Shi, Pengjian Shang, and Aijing Lin,
            "The coupling analysis of stock market indices based on 
            cross-permutation entropy."
            Nonlinear Dynamics
            79.4 (2015): 2439-2447.
     
   """

    assert len(Sig)<=2,  """Input arguments to be passed as data sequences:
        - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
        - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
    if len(Sig)==1:
        Sig = np.squeeze(Sig)
        assert max(Sig.shape)>10 and min(Sig.shape)==2,  """Input arguments to be passed as data sequences:
            - A single Nx2 numpy matrix with each column representing Sig1 and Sig2 respectively.       \n or \n
            - Two individual numpy vectors representing Sig1 and Sig2 respectively."""
        if Sig.shape[0] == 2:
            Sig = Sig.transpose()  
        S1 = Sig[:,0]; S2 = Sig[:,1]     
        
    elif len(Sig)==2:
        S1 = np.squeeze(Sig[0])
        S2 = np.squeeze(Sig[1])
        assert len(S1)==len(S2),  "Sig1/Sig2:   Each sequence must be of equal length"
        
    N1 = S1.shape[0]
    N2 = S2.shape[0]
    
    assert N1>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"
    assert isinstance(m,int) and (m > 2), "m:     must be an integer > 2"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
        
    N = N1-(m-1)*tau
    Sx1 = np.zeros((N,m))
    Sx2 = np.zeros((N,m))    
    for k in range(m):
        Sx1[:,k] = S1[k*tau:N+k*tau]
        Sx2[:,k] = S2[k*tau:N+k*tau]        
    
    Temp = np.argsort(Sx1,axis=1)
    Gx = np.zeros((N,m))
    for k in range(N):
        Gx[k,:] = Sx2[k,Temp[k,:]]
        
    Kt = np.zeros((m-2,m-2,N))
    for k in range(m-2):
        for j in range(k+1,m-1):
            G1 = Gx[:,j+1] - Gx[:,k]
            G2 = Gx[:,k] - Gx[:,j]
            Kt[k,j-1,:] = (G1*G2 > 0)
     
    Di = np.squeeze(np.sum(np.sum(Kt,1),0))
    Ppi,_ = np.histogram(Di,np.arange(-.5,((m-2)*(m-1)+2)/2))
    Ppi = Ppi[Ppi!=0]/N
    XPerm = -sum(Ppi*(np.log(Ppi)/np.log(Logx)))
    
    if round(sum(Ppi),6)!=1:
        print('Warning: Potential error with probability calculation')
        
    return XPerm
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base corrected Conditional Entropy function."""
import numpy as np

def CondEn(Sig, m=2, tau=1, c=6, Logx=np.exp(1), Norm=False):
    """CondEn  estimates the corrected conditional entropy of a univariate data sequence.

    .. code-block:: python
        
        Cond, SEw, SEz = CondEn(Sig) 
        
    Returns the corrected conditional entropy estimates (``Cond``) and the
    corresponding Shannon entropies (``m``: ``SEw``, ``m+1``: ``SEz``) for ``m`` = [1,2] 
    estimated from the data sequence (``Sig``) using the default parameters:
    embedding dimension = 2, time delay = 1, symbols = 6, logarithm = natural, normalisation = False
    Note: ``CondEn(m=1)`` returns the Shannon entropy of ``Sig``.
    
    .. code-block::
        
        Cond, SEw, SEz = CondEn(Sig, keyword = value, ...)
        
    Returns the corrected conditional entropy estimates (Cond) from the data
    sequence (Sig) using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, an integer > 1
        :tau:   - Time Delay, a positive integer
        :c:     - Number of symbols, an integer > 1
        :Logx:  - Logarithm base, a positive scalar 
        :Norm:  - Normalisation of ``Cond`` value, a boolean:
            * ``False``  no normalisation - default
            * ``True``   normalises w.r.t Shannon entropy of data sequence ``Sig``.  
 
    :See also:
        ``XCondEn``, ``MSEn``, ``PermEn``, ``DistEn``, ``XPermEn``
    
    :References:
        [1] Alberto Porta, et al.,
            "Measuring regularity by means of a corrected conditional
            entropy in sympathetic outflow." 
            Biological cybernetics 
            78.1 (1998): 71-78.
    """

    Sig = np.squeeze(Sig)    
        
    assert Sig.shape[0]>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 1), "m:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(c,int) and (c > 1), "c:   must be an integer > 1"
    assert isinstance(Logx,(int,float)) and Logx > 0, "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"
        
    Sig = (Sig-np.mean(Sig))/np.std(Sig)
    # Edges = np.linspace(min(Sig),max(Sig),c+1)
    Edges = np.arange(min(Sig),max(Sig),np.ptp(Sig)/c)
    Sx = np.digitize(Sig,Edges)
    # Sx[Sx==max(Sx)] = c
    N = Sx.shape[0]
    SEw = np.zeros(m-1)
    SEz = np.zeros(m-1)
    Prcm = np.zeros(m-1)
    Xi = np.zeros((N,m))
    Xi[:,-1] = Sx    
    for k in range(1,m):    #1:m-1
        Nx = N-(k*tau)
        Xi[:Nx,-(k+1)] = Sx[k*tau:]
        Wi = np.inner(c**np.arange(k-1,-1,-1),Xi[:Nx,-k:])
        Zi = np.inner(c**np.arange(k,-1,-1), Xi[:Nx,-(k+1):])
        Pw, _ = np.histogram(Wi,np.arange(min(Wi)-.5,max(Wi)+1.5))
        Pz, _ = np.histogram(Zi,np.arange(min(Zi)-.5,max(Zi)+1.5))
        Prcm[k-1] = sum(Pw==1)/Nx
        
        if sum(Pw)!=Nx or sum(Pz)!=Nx:
            print('Warning: Potential error estimating probabilities.')
        Pw = Pw[Pw!=0]; Pw = Pw/N
        Pz = Pz[Pz!=0]; Pz = Pz/N
        SEw[k-1] = -sum(Pw*np.log(Pw)/np.log(Logx))
        SEz[k-1] = -sum(Pz*np.log(Pz)/np.log(Logx))
        del Pw, Pz, Wi, Zi
        
    Temp, _ = np.histogram(Sx,c)
    Temp = Temp[Temp!=0]/N
    S1 = -sum(Temp*np.log(Temp)/np.log(Logx))
    Cond = SEz - SEw + Prcm*S1
    Cond = np.hstack((S1, Cond))
    if Norm:
        Cond = Cond/S1
    
    return Cond, SEw, SEz
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""